#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sun Jul 22 12:17:17 2018

@author: Sven Serneels, Ponalytics
"""

__name__ = "direpack"
__author__ = "Sven Serneels"
__license__ = "MIT"
__version__ = "1.0.19"
__date__ = "2021-09-13"

# The commented lines can be uncommented if IPOPT has been installed independently.  

from .preprocessing.robcent import VersatileScaler, versatile_scale
from .preprocessing.gsspp import GenSpatialSignPreProcessor, gen_ss_pp, gen_ss_covmat
from .sprm.sprm import sprm
from .sprm.snipls import snipls
from .sprm.rm import rm
from .cross_validation._cv_support_functions import robust_loss
from .ppdire.ppdire import ppdire
from .ppdire.capi import capi
from .dicomo.dicomo import dicomo
from .sudire.sudire import sudire, estimate_structural_dim
from .plot.sudire_plot import sudire_plot
from .plot.ppdire_plot import ppdire_plot
from .plot.sprm_plot import sprm_plot,sprm_plot_cv
from .ipopt_temp.ipopt_wrapper import minimize_ipopt
from .ipopt_temp.jacobian import *








#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Jul 9 14:20:17 2019

@author: Sven Serneels, Ponalytics
"""

__name__ = "dicomo"
__author__ = "Sven Serneels"
__license__ = "MIT"
__version__ = "1.0.3"
__date__ = "2021-04-07"









# -*- coding: utf-8 -*-
"""
Created on Sat Apr 11 17:25:42 2020

@author: Emmanuel Jordy Menvouta & Sven Serneels
"""

from __future__ import absolute_import, division, print_function
from __future__ import unicode_literals



from ..ppdire.ppdire import ppdire
from ..utils.utils import MyException
import matplotlib.pyplot as pp 
import numpy as np


class ppdire_plot(ppdire):
    
    def __init__(self,res_ppdire,colors,markers=['o','d','v'],*args):
        """
        Initialize with 
        res_ppdire, a ppdire class object
        
        Only mandatory input is colors, a list of colors for 
            [0] borders of pane 
            [1] plot background
            [2] marker fill
            [3] diagonal line 
            [4] marker contour, if different from fill
            [5] marker color for new cases, if applicable
               
        """
        if not(isinstance(res_ppdire,ppdire)):
            raise(MyException("Object supplied to ppdireplot needs to be a ppdire object"))
        self.res_ppdire = res_ppdire
        self.colors = colors
        self.markers = markers
        
    def plot_yyp(self,ytruev=[],Xn=[],label=[],names=[],namesv=[],title=[],legend_pos='lower right',onlyval=False):
        """
        plot_yyp will plot y vs y predicted for ppdire M opbjects
        Optional inputs: 
            ytruev: array (new_cases,) of predictands
            Xn: array (new_cases,variables) of predictors 
            If these arguments are supplied, ppdire predictions for ytrue will be 
                made from Xn through res_ppdire.predict()
            label: string: name of variable to be plotted. Will show in legend.
            names: list or tuple of strings, casenames from training set
            namesv: list or tuple of strings, casenames from test set
            title: String containing plot title
            legend_pos: string containing legend position
            onlyval: boolean: only plot validation cases
        """
        
        if len(label)==0:
            label = 'none'
        fig = pp.figure()
        fig.set_facecolor(self.colors[0])
        pp.rcParams['axes.facecolor'] = self.colors[1]
        ax1 = fig.add_subplot(111)
        if (not(onlyval)):
            ytruec = self.res_ppdire.y0
            if len(ytruec.shape) >1:
                ytruec = np.array(ytruec).reshape(-1).astype('float64')
            ypredc = np.array(self.res_ppdire.fitted_).T.reshape(-1)
            labelcr = label + ' Training'
            ax1.scatter(ytruec, ypredc, c=self.colors[2], label=labelcr,
                    zorder=1,edgecolors=self.colors[4],marker=self.markers[0])
            pp.xlabel("y-true")
            pp.ylabel("y-pred")
            
        else:
            if (len(Xn)==0):
                ValueError('In onlyval=True mode, new cases Xn need to be provided')
        if not(len(Xn)==0):
            if len(ytruev.shape) >1:
                ytruev = np.array(ytruev).reshape(-1).astype('float64')
            ypredv = self.res_ppdire.predict(Xn)
            ypredv = np.array(ypredv).reshape(-1).astype('float64')
            labelvr = label + ' Test'   
            ax1.scatter(ytruev,ypredv,c=self.colors[5],label=labelvr,
                        zorder=1,edgecolors=self.colors[4],marker=self.markers[0])
            pp.xlabel("y-true")
            pp.ylabel("y-pred")
           
        x_abline = np.array(ax1.get_xbound())
        ax1.add_line(pp.Line2D(x_abline,x_abline,color=self.colors[3]))
        if len(label)==0:
            ax1.legend_.remove()
        else:
            pp.legend(loc=legend_pos)
        if len(names)>0:
            if not(onlyval):
                for i in range(0,len(names)-1):
                    ax1.annotate(names[i], (ytruec[i],ypredc[i]))
        if len(namesv)>0:
            for i in range(0,len(namesv)-1):
                ax1.annotate(namesv[i], (ytruev[i],ypredv[i]))
        if len(title)>0:
            pp.title(title)
        pp.show()
        
    def plot_projections(self,Xn=[],label=[],components = [0,1],names=[],namesv=[],title=[],legend_pos='lower right',onlyval=False):
        
        """
        plot_projections will plot the score space  
        Optional inputs: 
            Xn: array (new_cases,variables) of predictors 
            If supplied, ppdire projections for new cases will be 
                made from Xn through res_ppdire.transform()
            label: string: name of variable to be plotted. Will show in legend.
            names: list or tuple of strings, casenames from training set
            namesv: list or tuple of strings, casenames from test set
            title: String containing plot title
            legend_pos: string containing legend position
            onlyval: boolean: only plot validation cases
        """
        
        if len(label)==0:
            label = 'none'
        fig = pp.figure()
        fig.set_facecolor(self.colors[0])
        pp.rcParams['axes.facecolor'] = self.colors[1]
        ax1 = fig.add_subplot(111)
        if (not(onlyval)):
            Tc = np.array(self.res_ppdire.x_scores_)
            labelcr = label + ' Training'
            ax1.scatter(Tc[:,components[0]], Tc[:,components[1]], c=self.colors[2], label=labelcr, 
                    zorder=1,edgecolors=self.colors[4],marker=self.markers[0])
        else:
            if (len(Xn)==0):
                ValueError('In onlyval=True mode, new cases Xn need to be provided')
        if not(len(Xn)==0):
            Tv = np.array(self.res_ppdire.transform(Xn))
            labelvr = label + ' Test'  
            ax1.scatter(Tv[:,components[0]], Tv[:,components[1]],c=self.colors[5],label=labelvr,
                        zorder=1,edgecolors=self.colors[4],marker=self.markers[0])
        if len(label)==0:
            ax1.legend_.remove()
        else:
            pp.legend(loc=legend_pos)
        if len(names)>0:
            if not(onlyval):
                for i in range(0,len(names)-1):
                    ax1.annotate(names[i], (Tc[i,components[0]], Tc[i,components[1]]))
        if len(namesv)>0:
            for i in range(0,len(namesv)-1):
                ax1.annotate(namesv[i], (Tv[i,components[0]], Tv[i,components[1]]))
        if len(title)>0:
            pp.title(title)
        pp.show()
        
        
        




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Apr 13 16:08:22 2020

@author: sven
"""

import pandas as ps
import numpy as np

class MyException(Exception):
        pass

def convert_X_input(X):
    
    if type(X) == ps.core.frame.DataFrame:
            X = X.to_numpy().astype('float64')
    return(X)
            
def convert_y_input(y):

    if type(y) in [ps.core.frame.DataFrame,ps.core.series.Series]:
        y = y.to_numpy().T.astype('float64')
    return(y)


def const_xscale(beta, *args):
    X= args[0]
    h= args[1]
    i = args[2]
    j = args[3]
    beta = np.reshape(beta,(-1,h),order = 'F')
    covx = np.cov(X, rowvar=False)
    ans = np.matmul(np.matmul(beta.T,covx), beta) - np.identity(h)
    return(ans[i,j])

def const_zscale(beta, *args):
    X= args[0]
    h= args[1]
    i = args[2]
    j = args[3]
    beta = np.reshape(beta,(-1,h),order = 'F')
    covx = np.identity(X.shape[1])
    ans = np.matmul(np.matmul(beta.T,covx), beta) - np.identity(h)
    return(ans[i,j])

def _predict_check_input(Xn):
    if type(Xn) == ps.core.series.Series:
        Xn = Xn.to_numpy()
    if Xn.ndim==1:
        Xn = Xn.reshape((1,-1))
    if type(Xn) == ps.core.frame.DataFrame:
        Xn = Xn.to_numpy()
    n,p = Xn.shape
    return (n,p,Xn)

def _check_input(X): 
    
    if(type(X) in (np.matrix,ps.core.frame.DataFrame,ps.core.series.Series)): 
        X = np.array(X)
        
    if (X.dtype == np.dtype('O')):
        X = X.astype('float64')
    
    if X.ndim == 1: 
        X = X.reshape((1,-1))
    
    n,p = X.shape 
    
    if n==1:
        if p >= 2: 
            X = X.reshape((-1,1))
    return(X)


# -*- coding: utf-8 -*-
"""
Created on Sat Apr 11 17:22:09 2020

@author: Emmanuel Jordy Menvouta
Edits by Sven Serneels. 
"""



__name__ = "sudire"
__author__ = "Emmanuel Jordy Menvouta"
__license__ = "MIT"
__version__ = "0.1.4"
__date__ = "2021-05-20"



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#@author: Sven Serneels, Ponalytics
# Created on Sun Feb 4 2018
# Updated on Sun Dec 16 2018
# Refactored on Sat Dec 21 2019
# Refactored on Sat Mar 28 2020


# Class for classical and robust centering and scaling of input data for 
# regression and machine learning 

# Version 2.0: Code entirely restructured compared to version 1.0. 
# Code made consistent with sklearn logic: fit(data,params) yields results. 
# Code makes more effciient use of numpy builtin estimators.
# Version 3.0:
# Code now takes strings or functions as input to centring and scaling. 
# Utility functions have been moved to _preproc_utilities.py 
# Code now supplied for l1median cetring, with options to use different 
# scipy.optimize optimization algorithms
# Version 4.0: 
# Made the API compatible for ScikitLearn pipelines. However, some nonstandard 
# functions and output remain for backwards compatibility. Functionality for
# sparse matrices still has to be implemented.  


# Ancillary functions in _preproc_utilities.py:

#         -   `scale_data(X,m,s)`: centers and scales X on center m (as vector) and scale s (as vector).
#         -   `mean(X,trimming)`: Column-wise mean.
#         -   `median(X)`: Column-wise median.
#         -   `l1median(X)`: L1 or spatial median. Optional arguments: 
#         -   `x0`: starting point for optimization, defaults to column wise median  
#         -   `method`: optimization algorithm, defaults to 'SLSQP' 
#         -   `tol`: tolerance, defaults to 1e-8
#         -   `options`: list of options for `scipy.optimize.minimize`
#         -   `kstepLTS(X): k-step LTS estimator of location.
#         -   `maxit`: int, number of iterations to compute maximally 
#         -   `tol`: float, tolerance for convergence
#         -   `std(X,trimming)`: Column-wise std.
#         -   `mad(X,c)`: Column-wise median absolute deviation, with consistency factor c.
#         -   `scaleTau2(x0, c1 = 4.5, c2 = 3, consistency = True)`: Tau estimator of scale 
#             with consistency parameters c1 and c2 and option for consistency correction
#             (True, False or 'finiteSample')



from __future__ import absolute_import, division, print_function
from __future__ import unicode_literals

from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.utils.metaestimators import _BaseComposition
from sklearn.utils.validation import check_is_fitted
import numpy as np
from ..utils.utils import MyException, convert_X_input, convert_y_input, _check_input
from ._preproc_utilities import *
from ._preproc_utilities import _check_trimming

__all__ = ['VersatileScaler','robcent','versatile_scale']

class VersatileScaler(_BaseComposition,TransformerMixin,BaseEstimator):

    """
    VersatileScaler Center and Scale data about classical or robust location and scale estimates  
    
    Parameters
    ----------
        `center`: str or callable, location estimator. String has to be name of the 
                function to be used, or 'None'. 
        `scale`: str or callable, scale estimator
        `trimming`: trimming percentage to be used in location and scale estimation. 


    Attributes
    ----------   
    Arguments for methods: 
        -   `X`: array-like, n x p, the data.
        -   `trimming`: float, fraction to be trimmed (must be in (0,1)). 
        
    

                
    Remarks
    -------
    Options for classical estimators 'mean' and 'std' also give access to robust 
    trimmed versions.


"""
    
    def __init__(self,center='mean',scale='std',trimming=0):
        
        """
        Initialize values. Check if correct options provided. 
        """
        
        self.center = center
        self.scale = scale 
        self.trimming = trimming
        
    
    def fit(self,X):
        
        """
        Estimate location and scale, store these in the class object.  
        Trimming fraction can be provided as keyword argument.
        """
        
        X = _check_input(X)
        
        _check_trimming(self.trimming)
        
        if type(self.center) is str: 
            center = eval(self.center)
        else:
            center = self.center
            
        if type(self.scale) is str: 
            scale = eval(self.scale)
        else:
            scale = self.scale
            

        n = X.shape
        if len(n) > 1:
            p = n[1]
        else:
            p = 1
        n = n[0] 
            
        if self.center == "None":
            m = np.repeat(0,p)
        else:
            m = center(X,trimming=self.trimming)
         
        # Keeping col_loc_ for older version compatibility
        setattr(self,"col_loc_",m)
        # sklearn standard 
        setattr(self,"center_",m)    
            
        if self.scale == "None":
            s = np.repeat(1,p)
        else:
            s = scale(X,trimming=self.trimming)
        
        # Keeping col_sca_ for older version compatibility
        setattr(self,"col_sca_",s)
        # sklearn standard 
        setattr(self,"scale_",s)
        
    
    def transform(self,X):
        
        """
        Center and/or scale training data to pre-estimated location and scale
        """
        
        X = _check_input(X)
        check_is_fitted(self,['center_','scale_'])
        
        Xs = scale_data(X,self.center_,self.scale_)
        setattr(self,'datas_',Xs)
            
        return Xs
        
    
    def predict(self,Xn):
        
        """
        Standardize new data on previously estimated location and scale. 
        Number of columns needs to match.
        """
        
        Xn = _check_input(Xn)
        Xns = scale_data(Xn,self.col_loc_,self.col_sca_)
        setattr(self,'datans_',Xns)
        return(Xns)
        
    def fit_transform(self,X):
        
        """
        Estimate center and scale for training data and scale these data 
        """
        
        self.fit(X)
        self.transform(X)
        return(self.datas_)
        
    def inverse_transform(self,Xs=None):
        
        """
        Transform scaled data back to their original scale  
        """
        
        check_is_fitted(self,['center_','scale_'])
        if Xs is not None:
            Xs = _check_input(Xs)
        else:
            Xs = self.datas_
        return(np.multiply(Xs,self.scale_) + self.center_)
        
        
# For backwards compatibility
robcent = VersatileScaler
            
        
def versatile_scale(X,center='l1median',scale='mad',trimming=0):
    
    """
    Wrapper to scale based on present robcent implementation that uses 
    `fit` instead of `transform`
    """
    
    rc = VersatileScaler(center=center,scale=scale,trimming=trimming)
    return(rc.fit_transform(X))
        


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sun Jul 22 12:17:17 2018

@author: Sven Serneels, Ponalytics
"""

__name__ = "plot"
__author__ = "Sven Serneels"
__license__ = "MIT"
__version__ = "0.9.0"
__date__ = "2020-04-18"







#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Jan 24 2019

Module containing:
    
    Estimators
    ----------
    Robust M Regression (RM)

Depends on robcent class for robustly centering and scaling data, as well as on
the functions in _m_support_functions. 

@author: Sven Serneels, Ponalytics
"""
from __future__ import absolute_import, division, print_function
from __future__ import unicode_literals
from sklearn.base import RegressorMixin,BaseEstimator
from sklearn.utils.metaestimators import _BaseComposition
import copy
import numpy as np
import pandas as ps
from scipy.stats import norm, chi2
from ..preprocessing.robcent import VersatileScaler
from ..utils.utils import MyException,_predict_check_input, _check_input
from ._m_support_functions import *

class rm(_BaseComposition,BaseEstimator,RegressorMixin):
    
    """
    Robust M Regression 
    
    Parameters:
    -----------
    fun: str, downweighting function. 'Hampel' (recommended), 'Fair' or 
                'Huber'
    probp1: float, probability cutoff for start of downweighting 
                 (e.g. 0.95)
    probp2: float, probability cutoff for start of steep downweighting 
                 (e.g. 0.975, only relevant if fun='Hampel')
    probp3: float, probability cutoff for start of outlier omission 
                 (e.g. 0.999, only relevant if fun='Hampel')
    centre: str, type of centring (`'mean'`, `'median'` or `'l1median'`, 
            the latter recommended statistically, if too slow, switch to `'median'`)
    scale: str, type of scaling ('std','mad' [recommended] or 'None')
    verbose: boolean, specifying verbose mode
    maxit: int, maximal number of iterations in M algorithm
    tol: float, tolerance for convergence in M algorithm 
    start_cutoff_mode: str, values:
        'specific' will set starting value cutoffs specific to X and y (preferred); 
        any other value will set X and y stating cutoffs identically. 
        The latter yields identical results to the SPRM R implementation available from
        CRAN. 
    copy (def True): boolean, whether to copy data
        Note: copy not yet aligned with sklearn def  
    
    """
    
    def __init__(self,fun='Hampel',probp1=0.95
                 ,probp2=0.975,probp3=0.999,centre='median',scale='mad'
                 ,start_cutoff_mode='specific',verbose=True,maxit=100
                 ,tol=0.01,copy=True):
        self.fun = fun
        self.probp1 = probp1
        self.probp2 = probp2
        self.probp3 = probp3
        self.centre = centre
        self.scale = scale
        self.start_cutoff_mode = start_cutoff_mode
        self.verbose = verbose
        self.maxit = maxit
        self.tol = tol
        self.copy = copy
        self.probctx_ = 'irrelevant'
        self.probcty_ = 'irrelevant'
        self.hampelbx_ = 'irrelevant'
        self.hampelby__ = 'irrelevant'
        self.hampelrx_ = 'irrelevant'
        self.hampelry_ = 'irrelevant'

    def fit(self,X,y):
        if self.copy:
            self.X = copy.deepcopy(X)
            self.y = copy.deepcopy(y)
        (n,p) = X.shape
        if (not(self.fun in ("Hampel", "Huber", "Fair"))):
            raise MyException("Invalid weighting function. Choose Hampel, Huber or Fair for parameter fun.")
        if ((self.probp1 > 1) | (self.probp1 <= 0)):
            raise MyException("probp1 is a probability. Choose a value between 0 and 1")
        if (self.fun == "Hampel"):
            if (not((self.probp1 < self.probp2) & (self.probp2 < self.probp3) & (self.probp3 <= 1))):
                raise MyException("Wrong choise of parameters for Hampel function. Use 0<probp1<hampelp2<hampelp3<=1")
            
        if type(X) == ps.core.frame.DataFrame:
            X = X.to_numpy()
        if type(y) in [ps.core.frame.DataFrame,ps.core.series.Series]:
            y = y.to_numpy().astype('float64')
        X = _check_input(X)
        y = _check_input(y)
        if len(y.shape) >1:
            y = np.array(y).reshape(-1).astype('float64')
        ny = y.shape[0]
        if ny != n:
            raise MyException("Number of cases in y and X must be identical.")
        
        scaling = VersatileScaler(center=self.centre, scale=self.scale)
        Xs = scaling.fit_transform(X).astype('float64')
        mX = scaling.col_loc_
        sX = scaling.col_sca_
        ys = scaling.fit_transform(y).astype('float64')
        my = scaling.col_loc_
        sy = scaling.col_sca_
        ys = np.array(ys).reshape(-1)

        wx = np.sqrt(np.array(np.sum(np.square(Xs),1),dtype=np.float64))
        wx = wx/np.median(wx)
        if [self.centre,self.scale]==['median','mad']:
            wy = np.array(abs(ys),dtype=np.float64)
        else:
            wy = (y - np.median(y))/(1.4826*np.median(abs(y-np.median(y))))
        self.probcty_ = norm.ppf(self.probp1)
        if self.start_cutoff_mode == 'specific':
            self.probctx_ = chi2.ppf(self.probp1,p)
        else: 
            self.probctx_ = self.probcty_
        if (self.fun == "Fair"):
            wx = Fair(wx,self.probctx_)
            wy = Fair(wy,self.probcty_)
        if (self.fun == "Huber"):
            wx = Huber(wx,self.probctx_)
            wy = Huber(wy,self.probcty_)
        if (self.fun == "Hampel"):
            self.hampelby_ = norm.ppf(self.probp2)
            self.hampelry_ = norm.ppf(self.probp3)
            if self.start_cutoff_mode == 'specific':
                self.hampelbx_ = chi2.ppf(self.probp2,p)
                self.hampelrx_ = chi2.ppf(self.probp3,p)
            else: 
                self.hampelbx_ = self.hampelby_
                self.hampelrx_ = self.hampelry_
            wx = Hampel(wx,self.probctx_,self.hampelbx_,self.hampelrx_)
            wy = Hampel(wy,self.probcty_,self.hampelby_,self.hampelry_)
        wx = np.array(wx).reshape(-1)
        w = (wx*wy).astype("float64")
        if (w < 1e-06).any():
            w0 = np.where(w < 1e-06)[0]
            w[w0] = 1e-06
            we = np.array(w,dtype=np.float64)
        else:
            we = np.array(w,dtype=np.float64)
        wye = wy
        WEmat = np.array([np.sqrt(we) for i in range(1,p+1)],ndmin=1).T    
        Xw = np.multiply(Xs,WEmat).astype("float64")
        yw = ys*np.sqrt(we)
        loops = 1
        rold = 1E-5
        difference = 1
        
        while ((difference > self.tol) & (loops < self.maxit)):
            b = np.linalg.lstsq(Xw,yw,rcond=None)
            b = np.array(b[0]).reshape(-1,1)
            yp = np.dot(Xs,b).reshape(-1)
            r = ys - yp
            if (len(r)/2 > np.sum(r == 0)):
                r = abs(r)/(1.4826 * np.median(abs(r)))
            else:
                r = abs(r)/(1.4826 * np.median(abs(r[r != 0])))
            wye = r
            if (self.fun == "Fair"):
                wye = Fair(wye,self.probcty_)
            if (self.fun == "Huber"):
                wye = Huber(wye,self.probcty_)
            if (self.fun == "Hampel"):
                wye = Hampel(wye,self.probcty_,self.hampelby_,self.hampelry_)
            b2sum = np.sum(np.square(b)) 
            difference = abs(b2sum - rold)/rold
            rold = b2sum
            we = (wye * wx).astype("float64")
            w0=[]
            if (any(we < 1e-06)):
                w0 = np.where(we < 1e-06)[0]
                we[w0] = 1e-06
                we = np.array(we,dtype=np.float64)
            if (len(w0) >= (n/2)):
                break
            WEmat = np.array([np.sqrt(we) for i in range(1,p+1)],ndmin=1).T    
            Xw = np.multiply(Xs,WEmat).astype("float64")
            yw = ys*np.sqrt(we)
            loops += 1
        if (difference > self.maxit):
            print("Warning: Method did not converge. The scaled difference between norms of the coefficient vectors is " + 
                  str(round(difference,4)))
        plotprec = False
        if plotprec:
            print(str(loops - 1))
        w = we
        w[w0] = 0
        wx[w0] = 0
        wy = wye
        wy[w0] = 0
        Xrw = np.array(np.multiply(Xs,np.sqrt(WEmat)).astype("float64"))
        scaling.set_params(scale='None')
        Xrw = scaling.fit_transform(Xrw) 
        b_rescaled = np.multiply(np.reshape(sy/sX,(p,1)),b)
        yp_rescaled = np.matmul(X,b_rescaled).reshape(-1)
        if(self.centre == "mean"):
            intercept = np.mean(y - yp_rescaled)
        else:
            intercept = np.median(y - yp_rescaled)
        yfit = yp_rescaled + intercept   
        if (self.scale!="None"):
            if (self.centre == "mean"):
                b0 = np.mean(ys.astype("float64") - np.matmul(Xs.astype("float64"),b))
            else:
                b0 = np.median(np.array(ys.astype("float64") - np.matmul(Xs.astype("float64"),b)))
        else:
            if (self.centre == "mean"):
                ytil = np.array(np.matmul(X,b)).reshape(-1)
                intercept = np.mean(y - ytil)
            else:
                intercept = np.median(y - ytil)
            b0 = intercept
        r = y - yfit
        setattr(self,"coef_",b_rescaled)
        setattr(self,"intercept_",intercept)
        setattr(self,"coef_scaled_",b)
        setattr(self,"intercept_scaled_",b0)
        setattr(self,"residuals_",r)
        setattr(self,"fitted_",yfit)
        setattr(self,"x_caseweights_",wx)
        setattr(self,"y_caseweights_",wy)
        setattr(self,"caseweights_",w)
        setattr(self,"x_loc_",mX)
        setattr(self,"y_loc_",my)
        setattr(self,"x_sca_",sX)
        setattr(self,"y_sca_",sy)
        setattr(self,'scaling_',scaling)
        return(self)
        pass
    
        
    def predict(self,Xn):
        n,p,Xn = _predict_check_input(Xn)
        if p!= self.X.shape[1]:
            raise(ValueError('New data must have seame number of columns as the ones the model has been trained with'))
        return(np.matmul(Xn,self.coef_) + self.intercept_)
        



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sun Jan 27 11:42:23 2019

Ancillary tools for plotting sprm results. 
    
    # Deleted: ABLine2D class, was broken in Py 3.7
    cv_score_table (function): tranform sklearn GridSearchCV 
        results into Data Frame

@author: Sven Serneels
"""
import numpy as np
import pandas as ps
from sklearn.metrics import mean_squared_error
from scipy.stats import norm
from ..sprm._m_support_functions import Fair, Huber, Hampel
        
def cv_score_table(res_sprm_cv):
        
    """
    Internal function reorganizing sklearn GridSearchCV results to pandas table. 
    The function adds the cv score table to the object as cv_score_table_
    """
        
    n_settings = len(res_sprm_cv.cv_results_['params'])
    etas = [res_sprm_cv.cv_results_['params'][i]['eta'] for i in range(0,n_settings)]
    components = [res_sprm_cv.cv_results_['params'][i]['n_components'] for i in range(0,n_settings)]
    cv_score_table_ = ps.DataFrame({'etas':etas, 'n_components':components, 'score':res_sprm_cv.cv_results_['mean_test_score']})
    return(cv_score_table_)
    
def robust_loss(y,ypred,lfun=mean_squared_error,fun=Hampel,probct=norm.ppf(0.975),hampelb=norm.ppf(.99),hampelr=norm.ppf(.999)):
    
    """
    Weighted loss function to be used in sklearn cross-validation
    Inputs: 
        y: array or matrix, original predictand
        ypred, array or matrix, predicted values
        lfun, function: an sklearn loss metric that accepts caseweights, 
            e.g. sklearn.metrics.mean_squared_error
        fun: function, weight function, 
            e.g. Fair, Huber or Hampel from sprm.sprm._m_support_functions
        probct, hampelb, hampelr: float, cutoffs for weight functions
    Output:
        loss, float
    """            
    
    if len(ypred.shape) > 1:
        ypred = np.array(ypred).reshape(-1)
    ypred = ypred.astype('float64')
    if len(y.shape) > 1:
        y = np.array(y).reshape(-1)
    y = y.astype('float64')
    r = y - ypred
    w = fun(r,probct,hampelb,hampelr)
    return(lfun(y,ypred,sample_weight=w))
    


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sun Apr 12 2020

This folder is temporary. It copies a fix to ipopt: 
    https://github.com/matthias-k/optpy/blob/master/optpy/jacobian.py
    
Folder will stay in direpack until the latter has been released.

"""

__name__ = "opt_temp"
__author__ = "Sven Serneels"
__license__ = "MIT"
__version__ = "0.0.2"
__date__ = "2021-04-15"





# -*- coding: utf-8 -*-
"""
Created on Tue Jun 30 13:17:46 2020

@author: Emmanuel Jordy Menvouta
"""

import unittest
from ..dicomo.dicomo import dicomo
import pandas as ps
import numpy as np
import statsmodels.robust as srs
import scipy.stats as sps
import dcor as dc

class Testdicomo(unittest.TestCase):
    """ Test  methods in the dicomo class"""
    
    @classmethod
    def setUpClass(cls):
        print('...setupClass')
        
        
        
    @classmethod
    def tearDownClass(cls):
        print('...teardownClass')
        
        
    @classmethod    
    def setUp(self):
        self.data=ps.read_csv("./data/Returns_shares.csv")
        self.datav = np.matrix(self.data.values[:,2:8].astype('float64'))
        self.est = dicomo()
        self.x = self.datav[:,1]
        self.y = self.datav[:,0]
        self.n=self.data.shape[0]
        self.p = self.data.shape[1]
        
        
        
    @classmethod    
    def tearDown(self):
        del self.est
        del self.x
        del self.y
        del self.n
        del self.p
        
        
    
    def test_mom(self):
        """ Tests functions to compute moments"""
        
        self.assertAlmostEquals(self.est.fit(self.x,biascorr=False),np.var(self.x))# biased var
        self.assertAlmostEquals(self.est.fit(self.x,biascorr=True),np.var(self.x)*self.n/(self.n-1))#unbiased var
        self.est.set_params(center='median')
        self.assertAlmostEquals(self.est.fit(self.x),srs.mad(self.x)[0],places=4)
        self.est.set_params(center='mean')
        self.assertAlmostEquals(self.est.fit(self.x,biascorr=False,order=3),sps.moment(self.x,3)[0])#third moment
        self.est.set_params(mode='skew')
        self.assertAlmostEquals(self.est.fit(self.x,biascorr=False),sps.skew(self.x)[0])# skew without small sample corr
        self.assertAlmostEquals(self.est.fit(self.x,biascorr=True),sps.skew(self.x,bias=False)[0])
        
        
        
        
        
        
    def test_como(self):
        """ Tests function to compute comomennts"""
        
        self.est.set_params(mode='com')
        self.assertAlmostEquals(self.est.fit(self.x,y=self.y,biascorr=True),self.data.iloc[:,2:4].cov().values[0,1])#covariance
        self.assertAlmostEquals(self.est.fit(self.x,y=self.y,biascorr=True,option=1,order=3),0.39009,places=4)#third order comoment
        self.est.set_params(mode='corr')
        self.assertAlmostEquals(self.est.fit(self.x,y=self.y),self.data.iloc[:,2:4].corr().values[0,1])#correlation
        self.est.set_params(mode='continuum')
        self.assertAlmostEquals(np.sqrt(self.est.fit(self.x,y=self.y,alpha=1,biascorr=True)),self.data.iloc[:,2:4].cov().values[0,1])#continuum
        
        
    def test_energy(self):
        """ Tests function  to compute energy statistics"""
        
        self.est.set_params(est='distance',mode='var')
        self.assertAlmostEquals(self.est.fit(self.x,biascorr=False),dc.distance_stats(self.x,self.x)[0])
        self.assertAlmostEquals(self.est.fit(self.x,biascorr=True),np.sqrt(dc.u_distance_stats_sqr(self.x,self.x)[0]))      
        self.est.set_params(mode='com')
        self.assertAlmostEquals(self.est.fit(self.x,y=self.y,biascorr=False),dc.distance_covariance(self.x,self.y))
        self.est.set_params(mode='mdd')
        self.assertAlmostEquals(self.est.fit(self.x,y=self.y,biascorr=False),0.352427150086)
        
        
        
        
        
        

        
        
        

        
        
        
        
        
        
        
    
    
if __name__ =='__main__':
    unittest.main()
        
    


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sun Dec 16 13:44:19 2018

@author: sven
"""

import scipy.stats as sps
import scipy.spatial as spp
import numpy as np
import copy
from ..utils.utils import MyException

def trim_mean(x,trimming,axis=0):
    """
    computes the trimmed mean of array x according to axis.
    Input :
        x : input data as numpy array
        trimming, float : trimming percentage to be used
        axis, int or None : Axis along which the trimmed means are computed
        
    Output:
        The trimmed mean of x according to axis.
    
    """
    
    if trimming == 0: 
        return(np.mean(x,axis=axis))
    else:
        return(sps.trim_mean(x,trimming,axis=axis))

def trimvar(x,trimming):
    
    """
    computes the trimmed variance of array x .
    Input :
        x : input data as numpy array
        trimming, float : trimming percentage to be used
     
    Output:
        The trimmed variance of x.
    
    """
        # division by n
    return(sps.trim_mean(np.square(x - sps.trim_mean(x,trimming)),trimming))
        
def identity(x): 
    return(x)
        
def trim_mom(x,y,locest,order,trimming,option,fscorr=True):
    """
    computes trimmed comoment   between x and y. order represents the order of 
    the comoment.
    input :
        x : Input data as matrix 
        y : Input data as matrix or 1d vector
        order, int : order of the comoment
        trimming, float : trimming percentage to be used.
        option, int : option to select the type of co-moment (order 3: option 1 = com(x,x,y))
        fscor, bool: if True, a finite sample correction is applied to the comoment. 
    
    output : 
        the trimmed comoment between x and y 
    """
        # division by n
        
    if order == 0:
        como = 0
    elif order == 1: 
        como = locest(x,trimming)
    else:
        if order > 2: 
            iter_stop_2 = option 
            iter_stop_1 = order - option
        else: 
            iter_stop_1 = 1
            iter_stop_2 = 1
        
        if locest == np.median:
            trimming = 0
            factor = 1
            if (x==y).all():
                wrapper = abs
                power = 1/order
                if power == 0.5:
                    factor = 1.4826
            else:
                wrapper = identity
                power = 1
        else:
            n = len(x)
            wrapper = identity
            power = 1
            if fscorr:
                ntrim = round(n * (1-trimming)) 
                factor = ntrim
                factor /= np.product(ntrim - np.arange(max(1,order-2),order))
            else:
                factor = 1
    
        xc = wrapper(x - locest(x,trimming))
        yc = wrapper(y - locest(y,trimming))
    
        factor1 = np.power(xc,iter_stop_1)
        factor2 = np.power(yc,iter_stop_2)
    
        como = locest(np.power(np.multiply(factor1,factor2),power),trimming)*factor
#        como = sps.trim_mean(np.multiply(x - sps.trim_mean(x,trimming),y - sps.trim_mean(y,trimming)),trimming)*ntrim/(ntrim-1)
    if len(como.shape)>1: 
        como = como[0,0]
    else:
        if type(como) is np.ndarray:
            como = como[0]
            
        
    return(como)

def double_center_flex(a, center='mean', **kwargs):
    """
    Double centered function adapted to accommodate for location types different
    from mean.
    Input : 
        a : input data as matrix
        center, str : which location estimate to use for centering. either 'mean or 'median'
        kwargs :
            trimming, float : trimming percentage to be used.
            biascorr, bool : if True, bias correction is applied during double centering.
    Output : 
        The double centered version of the matrix a.

    """
    
    # print(kwargs)
    
    if 'trimming' not in kwargs:
        trimming = 0
    else:
        trimming = kwargs.get('trimming')
        # print('trimming is: ' + str(trimming))
        
    if 'biascorr' not in kwargs:
        biascorr = False
    else:
        biascorr = kwargs.get('biascorr')
    
    out = copy.deepcopy(a)
    
    dim = np.size(a, 0)
    n1 = dim

    # mu = np.sum(a) / (dim * dim)
    if center=='mean':
        mu = trim_mean(a.reshape((dim**2,1)),trimming)
        if biascorr:
            n1 = np.round(dim*(1-trimming))
            # print(n1)
            mu *= (n1**2) / ((n1-1) * (n1-2))
        mu_cols = trim_mean(a, trimming, axis=0).reshape((1,dim))
        mu_rows = trim_mean(a, trimming, axis=1).reshape((dim,1))
        if biascorr:
            mu_cols *= n1/(n1 - 2)
            mu_rows *= n1/(n1 - 2)
        mu_cols = np.ones((dim, 1)).dot(mu_cols)
        mu_rows = mu_rows.dot(np.ones((1, dim)))
    elif center=='median':
        mu = np.median(a.reshape((dim**2,1)))
        mu_cols = np.median(a,axis=0).reshape((1,dim))
        mu_rows = np.median(a,axis=1).reshape((dim,1))
        mu_cols = np.ones((dim, 1)).dot(mu_cols)
        mu_rows = mu_rows.dot(np.ones((1, dim)))
    else:
        raise(ValueError('Center should be mean or median'))
        

    # Do one operation at a time, to improve broadcasting memory usage.
    out -= mu_rows
    out -= mu_cols
    out += mu
    
    if biascorr:
        out[np.eye(dim, dtype=bool)] = 0

    return out,n1


def distance_matrix_centered(x,**kwargs):
    """
    Computes the trimmed double centered distance matrix of x.
    Input : 
        x : input data as matrix.
        kwargs :
            trimming, float : trimming percentage to be used.
            biascorr, bool : if True, bias correction is applied during double centering.
            center, str : which location estimate to use for centering. either 'mean or 'median'
            dmetric, str : which distance metric to use. Default is euclidean distance.
    Output :
        the trimmed double centered distance matrix of x            
    """
    
    if 'trimming' not in kwargs:
        trimming = 0
    else:
        trimming = kwargs.get('trimming')
            
    if 'biascorr' not in kwargs:
        biascorr = False
    else:
        biascorr = kwargs.get('biascorr')
        
    if 'center' not in kwargs:
        center = 'mean'
    else:
        center = kwargs.get('center')
        
    if 'dmetric' not in kwargs:
        dmetric = 'euclidean'
    else:
        dmetric = kwargs.get('dmetric')
        
    dx = spp.distance.squareform(spp.distance.pdist(x,metric=dmetric))
    dmx, n1 = double_center_flex(dx,biascorr=biascorr,
                                 trimming=trimming,center=center)
    
    return dmx,n1


def distance_moment(dmx,dmy,**kwargs):
    """
    Computes the trimmed distance comoment between x and y based on their distance matrices.
    Input : 
        dmx : distance matrix of x 
        dmy : distance matrix of y 
        
        kwargs :
            trimming, float : trimming percentage to be used.
            biascorr, bool : if True, bias correction is applied during double centering.
            center, str : which location estimate to use for centering. either 'mean or 'median'
            dmetric, str : which distance metric to use. Default is euclidean distance.
            order, int : order  of the comoment to be computed, default is 2 for covariance.
            option, int : option to be used during the computation. 
    Output :
        The trimmed  distance comoment  between x and y
    
    """
    
    if 'trimming' not in kwargs:
        trimming = 0
    else:
        trimming = kwargs.get('trimming')
            
    if 'biascorr' not in kwargs:
        biascorr = False
    else:
        biascorr = kwargs.get('biascorr')
        
    if 'center' not in kwargs:
        center = 'mean'
    else:
        center = kwargs.get('center')
        
    if 'order' not in kwargs:
        order = 2
    else:
        order = kwargs.get('order')
        
    if order > 2: 
        if 'option' not in kwargs:
            option = 1
        else:
            option = kwargs.get('option')
            
        iter_stop_2 = option 
        iter_stop_1 = order - option
    else:
        option = 0
        iter_stop_1 = 1
        iter_stop_2 = 1
        
    nx = dmx.shape[0]
    ny = dmy.shape[0]
    if nx!=ny:
        raise(ValueError)
        
        
    if biascorr: 
        
        if trimming == 0:
            n1 = nx
        elif 'n1' not in kwargs:
            raise(MyException('n1 needs to be provided when correcting for bias'))
        else:
            n1 = kwargs.get('n1')
        
        corr4bias = n1**2/(n1*(n1-3))
        
    else:
        corr4bias = 1
        
    if order>2:
        i = 1
        while i < iter_stop_1:  
            dmx *= dmx
            i += 1
        i = 1
        while i < iter_stop_2:  
            dmy *= dmy
            i += 1
        
        
    if center=='mean':
        moment = trim_mean((dmx*dmy).reshape((nx**2,1)),trimming)
        moment *= corr4bias
        moment = moment[0]
        moment = (-1)**order*abs(moment)**(1/order)
    elif center=='median':
        moment = np.median(dmx*dmy)
        
    return(moment)
    
    
def difference_divergence(X,Y,**kwargs):
    
    
    """
    This function computes the (U)Martingale Difference Divergence of Y given X.
    
    input :
    
        X : A  matrix or data frame, where rows represent samples, and columns represent variables.
        Y : The response variable or matrix.
        biascorr, bool : if True, uses U centering to produce an unbiased estimator of MDD
        
    output:
        returns the squared martingale difference divergence of Y given X.
       
    """
    
    if 'trimming' not in kwargs:
        trimming = 0
    else:
        trimming = kwargs.get('trimming')
        
    if 'biascorr' not in kwargs:
        biascorr = False
    else:
        biascorr = kwargs.get('biascorr')
    if 'center' not in kwargs:
        center = 'mean'
    else:
        center = kwargs.get('center')
        
    if 'dmetric' not in kwargs:
        dmetric = 'euclidean'
    else:
        dmetric = kwargs.get('dmetric')
    
   
    
    
    
    A, Adim = distance_matrix_centered(X,biascorr=biascorr,trimming=trimming,center=center)  
    dy=  spp.distance.squareform(spp.distance.pdist(Y.reshape(-1, 1),metric=dmetric)**2)
    B,Bdim = double_center_flex(0.5*dy,biascorr=biascorr,trimming=trimming,center=center)
    if biascorr:
        return(U_inner(A,B,trimming))
    else:
        return(D_inner(A,B,trimming))
       
        
def U_inner(X,Y,trimming=0):
    
    """
        Computes the inner product in the space of U centered matrices, between matrices X and Y.
        The matrices have to be  square matrices.
    
    """
        
    nx = X.shape[0]
    ny = Y.shape[0]
    
    if nx != ny:
        raise(MyException('Please feed x and y data of equal length'))
        
    #((1/(nx*(nx-3))) *(np.sum(arr)))
    arr= np.multiply(X,Y)
    arr=arr.flatten()
    lowercut = int(trimming * (nx**2))
    uppercut = (nx**2) - lowercut
    atmp = np.partition(arr, (lowercut, uppercut - 1), axis=0)
    sl = [slice(None)] * atmp.ndim
    sl[0] = slice(lowercut, uppercut)
    res= atmp[tuple(sl)]
    n = np.sqrt(len(res))
    return((1/(n*(n-3)))*np.sum(atmp[tuple(sl)], axis=0))


def D_inner(X,Y,trimming=0):
    
    """
    Computes the inner product in the space of D centered matrices, between Double centered matrices X and Y.
    The matrices have to be square matrices.
    
    """
        
    nx = X.shape[0]
    ny = Y.shape[0]
    
    if nx != ny:
        raise(MyException('Please feed x and y data of equal length'))
        
    #arr= (1/(nx*nx))*np.multiply(X,Y)
    arr= np.multiply(X,Y)
    arr=arr.flatten()
    lowercut = int(trimming * (nx**2))
    uppercut = (nx**2) - lowercut
    atmp = np.partition(arr, (lowercut, uppercut - 1), axis=0)
    sl = [slice(None)] * atmp.ndim
    sl[0] = slice(lowercut, uppercut)
    res= atmp[tuple(sl)]
    n = np.sqrt(len(res))
    return((1/(n*n))*np.sum(res, axis=0))

def MDDM(X,Y): 
    
    """Computes the MDDM(Y|X)
    for more details, see the article by Chung Eun Lee & Xiaofeng Shao;
    Martingale Difference Divergence Matrix and Its
    Application to Dimension Reduction for Stationary
    Multivariate Time Series;                                                 
    Journal of the American Statistical Association; 2018;521;
    216--229
    
    
    Input: 
    X  ---  ndarray of shape (n,p)
    Y  --- ndarray of shape(n,q)
    
    Output: 
    MDDM(Y|X)
    """

    n,q = Y.shape
    n,p = X.shape
    MDDM = np.zeros((q,q))
    Y_mean = np.mean(Y,axis=0).reshape(1,-1)
    Y_center = Y - np.matmul(np.ones((n,1)),Y_mean)
    
    for i in range(n):
        if(p==1):
            X_dist = np.abs(X[i]-X)
            
        else: 
            X_diff= (( X.T - np.vstack(X[i,:])).T)**2
            X_sum = np.sum(X_diff,axis=1)
            X_dist = np.sqrt(X_sum).reshape(-1,n)
            
        
        MDDM = MDDM + np.matmul(Y_center[i,:].reshape(q,-1), np.matmul(X_dist,Y_center))
        
    
    MDDM = (-MDDM)/(n**2)
    
    return(MDDM)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sun May 12 10:03:05 2019

@author: Sven Serneels, Ponalytics.
"""

from sklearn.base import BaseEstimator, defaultdict
from sklearn.utils.metaestimators import _BaseComposition
from collections import defaultdict
import inspect
from ..dicomo.dicomo import dicomo 
from ..dicomo._dicomo_utils import * 

class capi(_BaseComposition,BaseEstimator):
    
    """
    CAPI Co-moment analysis projection index
    
    The CAPI projection index to estimate generalized betas was first introduced
    in: 
    
    S. Serneels, Projection pursuit based generalized betas accounting for 
    higher order co-moment effects in financial market analysis,  in: 
    JSM Proceedings, Business and Economic Statistics Section. 
    Alexandria, VA: American Statistical Association, 2019, 3009-3035.
    
    Class arguments 
    
    max_degree, int: maxmimal degree of co-moments to be used. In [2,3,4]. 
    
    projection_index, class object: class used to calculate co-moments. 
        Written to work with dicomo class yet other plugins could be written.
    
    pi_arguments, dict: dict of arguments to pass on to projection_index
    
    weights, list of float: weights to used in linear combination of co-moments. 
    
    centring, bool
    
    scaling, bool whether to calculate CAPI based on scaled higher co-moments 
        (co-skewness, co-kurtosis) or raw higher co-moments
    
    options, either a list of co-moment options to be included, or 'all' (e.g.
    option=i calculates M3,i and M4,i etc.)
    
    After intializing the object, call object.fit(x,y,**kwargs) to evaluate. 
    CAPI takes no direct kwargs, yet passes all kwargs on to the fit method of
    the projection index. 
    
    """
    
    
    def __init__(self,
                 max_degree=2, 
                 projection_index=dicomo,
                 pi_arguments={},
                 weights = [1,1,1,-1,-1,-1],
                 centring=False,
                 scaling=True, 
                 options='all'):
        self.max_degree = max_degree
        self.projection_index = projection_index
        self.pi_arguments = pi_arguments
        self.weights = weights
        self.most = self.projection_index(**self.pi_arguments)
        self.scaling = scaling
        self.options = options
        self.capi_index_ = None
        if self.max_degree > 4:
            raise(ValueError('Maximal degree is 4.'))
        
        
    def fit(self,x,y,**kwargs):
        
        
        if self.scaling:
            order_kwargs =['cov','cos','cok']
        else: 
            order_kwargs =['com','com','com']

        if self.max_degree < 2: 
            raise(ValueError('capi not meaningful for max_degree < 2'))
        if self.options=='all':
            options = np.arange(1,4)
        else: 
            options = np.array(self.options,ndmin=1)
        moments = np.zeros(6)
        fit_arguments = {'order': 0, 'y': y}
        fit_arguments = {**kwargs,**fit_arguments}
        init_moment_calc = 2
        k = 0
        for i in range(init_moment_calc,self.max_degree+1):
            fit_arguments['order'] = i
            self.most.set_params(mode=order_kwargs[i-2])
            l = min(i-1,len(options))
            for j in options[np.arange(0,l)]:
                fit_arguments['option'] = j
                moments[i-3+j+k] = self.most.fit(x,**fit_arguments)
            if i==3:
                k += 1
        capi_index_ = np.dot(self.weights,moments)
        self.capi_index_ = capi_index_
        self.moments_ = moments
        return(capi_index_)
        
        
    @classmethod   
    def _get_param_names(cls):
        """Get parameter names for the estimator"""
        # fetch the constructor or the original constructor before
        # deprecation wrapping if any
        init = getattr(cls.__init__, 'deprecated_original', cls.__init__)
        if init is object.__init__:
            # No explicit constructor to introspect
            return []
    
        # introspect the constructor arguments to find the model parameters
        # to represent
        init_signature = inspect.signature(init)
        # Consider the constructor parameters excluding 'self'
        parameters = [p for p in init_signature.parameters.values()
                      if p.name != 'self' and p.kind != p.VAR_KEYWORD]
        for p in parameters:
            if p.kind == p.VAR_POSITIONAL:
                raise RuntimeError("scikit-learn estimators should always "
                                   "specify their parameters in the signature"
                                   " of their __init__ (no varargs)."
                                   " %s with constructor %s doesn't "
                                   " follow this convention."
                                   % (cls, init_signature))
        # Extract and sort argument names excluding 'self'
        return sorted([p.name for p in parameters])
    
    def get_params(self, deep=False):
        """Get parameters for this estimator.
        Parameters
        ----------
        deep : boolean, optional
            If True, will return the parameters for this estimator and
            contained subobjects that are estimators.
        Returns
        -------
        params : mapping of string to any
            Parameter names mapped to their values.
        ------
        Copied from ScikitLlearn instead of imported to avoid 'deep=True'
        """
        out = dict()
        for key in self._get_param_names():
            value = getattr(self, key, None)
            if deep and hasattr(value, 'get_params'):
                deep_items = value.get_params().items()
                out.update((key + '__' + k, val) for k, val in deep_items)
            out[key] = value
        return out
        
    def set_params(self, **params):
        """Set the parameters of this estimator.
        Copied from ScikitLearn, adapted to avoid calling 'deep=True'
        Returns
        -------
        self
        ------
        Copied from ScikitLlearn instead of imported to avoid 'deep=True'
        """
        if not params:
            # Simple optimization to gain speed (inspect is slow)
            return self
        valid_params = self.get_params()
    
        nested_params = defaultdict(dict)  # grouped by prefix
        for key, value in params.items():
            key, delim, sub_key = key.partition('__')
            if key not in valid_params:
                raise ValueError('Invalid parameter %s for estimator %s. '
                                 'Check the list of available parameters '
                                 'with `estimator.get_params().keys()`.' %
                                 (key, self))
    
            if delim:
                nested_params[key][sub_key] = value
            else:
                setattr(self, key, value)
                valid_params[key] = value
    
        for key, sub_params in nested_params.items():
            valid_params[key].set_params(**sub_params)
    
        return self
    
            
        
        


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Created on Fri Apr 26 19:27:52 2019

# @author: sven


from __future__ import absolute_import, division, print_function
from __future__ import unicode_literals
from sklearn.base import RegressorMixin,BaseEstimator,TransformerMixin
from sklearn.utils.metaestimators import _BaseComposition
import copy
import numpy as np
import pandas as ps
from ..preprocessing.robcent import VersatileScaler
from ..utils.utils import MyException, _predict_check_input, _check_input
from ..preprocessing._preproc_utilities import scale_data 

class snipls(_BaseComposition,BaseEstimator,TransformerMixin,RegressorMixin):
    """
    SNIPLS Sparse Nipals Algorithm 
    
    Algorithm first outlined in: 
        Sparse and robust PLS for binary classification, 
        I. Hoffmann, P. Filzmoser, S. Serneels, K. Varmuza, 
        Journal of Chemometrics, 30 (2016), 153-162.
    
    Parameters
    -----------

    eta : float.
         Sparsity parameter in [0,1)

    n_components : int,
                     min 1. Note that if applied on data, n_components shall take a value <= min(x_data.shape)

    verbose: Boolean (def true)
                to print intermediate set of columns retained

    columns : Either boolean, list, numpy array or pandas Index (def false)
                if False, no column names supplied; if True, if X data are supplied as a pandas data frame, will extract column names from the frame throws an error for other data input types if a list, array or Index (will only take length x_data.shape[1]), the column names of the x_data supplied in this list, will be printed in verbose mode.

    centre : str, 
                type of centring (`'mean'` [recommended], `'median'` or `'l1median'`), 

    scale : str,
             type of scaling ('std','mad' or 'None')

    copy : (def True): boolean,
             whether to copy data.  Note : copy not yet aligned with sklearn def  - we always copy  
    
             
    Attributes
    ------------
    Attributes always provided:

        -  `x_weights_`: X block PLS weighting vectors (usually denoted W)
        -  `x_loadings_`: X block PLS loading vectors (usually denoted P)
        -  `C_`: vector of inner relationship between response and latent variablesblock re
        -  `x_scores_`: X block PLS score vectors (usually denoted T)
        -  `coef_`: vector of regression coefficients 
        -  `intercept_`: intercept
        -  `coef_scaled_`: vector of scaled regression coeeficients (when scaling option used)
        -  `intercept_scaled_`: scaled intercept
        -  `residuals_`: vector of regression residuals
        -  `x_ev_`: X block explained variance per component
        -  `y_ev_`: y block explained variance 
        -  `fitted_`: fitted response
        -  `x_Rweights_`: X block SIMPLS style weighting vectors (usually denoted R)
        -  `colret_`: names of variables retained in the sparse model
        -  `x_loc_`: X block location estimate 
        -  `y_loc_`: y location estimate
        -  `x_sca_`: X block scale estimate
        -  `y_sca_`: y scale estimate
        -  `centring_`: scaling object used internally (from `VersatileScaler`)
    
    """
    
    def __init__(self,eta=.5,n_components=1,verbose=True,columns=False,
                 centre='mean',scale='None',copy=True):
        self.eta = eta 
        self.n_components = n_components
        self.verbose = verbose
        self.columns = columns
        self.centre = centre
        self.scale = scale
        self.copy = copy

    def fit(self,X,y):
        """
            Fit a  SNIPLS model. 
            
            Parameters
            ------------ 
                
                X : numpy array 
                    Input data.

                y :   vector or 1D matrix
                    Response data

        """
        if type(self.columns) is list: 
            self.columns = np.array(self.columns)
        elif type(self.columns) is bool:
            if type(X) != ps.core.frame.DataFrame and self.columns:
                raise(MyException("Columns set to true can only extract column names for data frame input"))
        if type(X) == ps.core.frame.DataFrame:
            if type(self.columns) is bool and self.columns:
                self.columns = X.columns
            X = X.to_numpy()
        (n,p) = X.shape
        if type(y) in [ps.core.frame.DataFrame,ps.core.series.Series]:
            y = y.to_numpy()
        X = _check_input(X)
        y = _check_input(y)
        ny = y.shape[0]
        if ny != n:
            if y.ndim == 2:
                y = y.T
            else:
                raise(MyException("Number of cases in X and y needs to agree"))
        y = y.astype("float64")
        if self.copy:
            X0 = copy.deepcopy(X)
            y0 = copy.deepcopy(y)
        else:
            X0 = X
            y0 = y
        self.X = X0
        self.y = y0
        X0 = X0.astype("float64")
        centring = VersatileScaler(center=self.centre,scale=self.scale)
        X0 = centring.fit_transform(X0).astype('float64')
        mX = centring.col_loc_
        sX = centring.col_sca_
        y0 = centring.fit_transform(y0).astype('float64')
        my = centring.col_loc_
        sy = centring.col_sca_
        T = np.empty((n,self.n_components),float) 
        W = np.empty((p,self.n_components),float)  
        P = np.empty((p,self.n_components),float)
        C = np.empty((self.n_components,1),float) 
        Xev = np.empty((self.n_components,1),float)
        yev = np.empty((self.n_components,1),float)
        B = np.empty((p,1),float) 
        oldgoodies = np.array([])
        Xi = X0
        yi = y0
        for i in range(1,self.n_components+1):
            wh =  np.dot(Xi.T,yi)
            wh = wh/np.linalg.norm(wh,"fro")
            # goodies = abs(wh)-llambda/2 lambda definition
            goodies = abs(wh)-self.eta*max(abs(wh))
            wh = np.multiply(goodies,np.sign(wh))
            goodies = np.where((goodies>0))[0]
            goodies = np.union1d(oldgoodies,goodies)
            oldgoodies = goodies
            if len(goodies)==0:
                print("No variables retained at" + str(i) + "latent variables" +
                      "and lambda = " + str(self.eta) + ", try lower lambda")
                break
            elimvars = np.setdiff1d(range(0,p),goodies)
            wh[elimvars] = 0 
            th = np.dot(Xi,wh)
            nth = np.linalg.norm(th,"fro")
            ch = np.dot(yi.T,th)/(nth**2)
            ph = np.dot(Xi.T,np.dot(Xi,wh))/(nth**2)
            Xi = Xi - np.dot(th,ph.T)
            yi = yi - np.dot(th,ch)
            ph[elimvars] = 0 
            W[:,i-1] = np.reshape(wh,p)
            P[:,i-1] = np.reshape(ph,p)
            C[i-1] = ch 
            T[:,i-1] = np.reshape(th,n)
            Xev[i-1] = (nth**2*np.linalg.norm(ph,"fro")**2)/np.sum(np.square(X0))*100
            yev[i-1] = np.sum(nth**2*(ch**2))/np.sum(np.power(y0,2))*100
            if type(self.columns)==bool:
                colret = goodies
            else:
                colret = self.columns[np.setdiff1d(range(0,p),elimvars)]
            if(self.verbose):
                print("Variables retained for " + str(i) + " latent variable(s):" +
                      "\n" + str(colret) + ".\n")
        if(len(goodies)>0):
            R = np.matmul(W[:,range(0,i)] , np.linalg.inv(np.matmul(P[:,range(0,i)].T,W[:,range(0,i)])))
            B = np.matmul(W[:,range(0,i)],np.matmul(
                    np.linalg.inv(np.matmul(np.matmul(W[:,range(0,i)].T,
                                np.matmul(X0.T,X0)),W[:,range(0,i)])),
                    np.matmul(np.matmul(W[:,range(0,i)].T,X0.T),y0)))
        else:  
            B = np.empty((p,1))
            B.fill(0)
            R = B
            T = np.empty((n,self.n_components))
            T.fill(0)
        B_rescaled = np.multiply(np.array(sy/sX).reshape((p,1)),B)
        yp_rescaled = np.dot(X,B_rescaled)
        if(self.centre == "mean"):
            intercept = np.mean(y - yp_rescaled)
        else:
            intercept = np.median(y - yp_rescaled)
        yfit = yp_rescaled + intercept    
        yfit = yfit.reshape(-1)    
        r = y - yfit
        setattr(self,"x_weights_",W)
        setattr(self,"x_loadings_",P)
        setattr(self,"C_",C)
        setattr(self,"x_scores_",T)
        setattr(self,"coef_",B_rescaled)
        setattr(self,"coef_scaled_",B)
        setattr(self,"intercept_",intercept)
        setattr(self,"x_ev_",Xev)
        setattr(self,"y_ev_",yev)
        setattr(self,"fitted_",yfit)
        setattr(self,"residuals_",r)
        setattr(self,"x_Rweights_",R)
        setattr(self,"colret_",colret)
        setattr(self,"x_loc_",mX)
        setattr(self,"y_loc_",my)
        setattr(self,"x_sca_",sX)
        setattr(self,"y_sca_",sy)
        setattr(self,"centring_",centring)
        return(self)
        
    
    def predict(self,Xn):
        """
        Predict using a  SNIPLS model. 
        
        Parameters
        ------------ 
            
            Xn : numpy array or data frame 
                Input data.

        """
        n,p,Xn = _predict_check_input(Xn)
        if p!= self.X.shape[1]:
            raise(ValueError('New data must have same number of columns as the ones the model has been trained with'))
        return(np.matmul(Xn,self.coef_) + self.intercept_)
        
    def transform(self,Xn):
        """
        Transform input data. 
        
        
        Parameters
        ------------ 
            
            Xn : numpy array or data frame 
                Input data.

        """
        n,p,Xn = _predict_check_input(Xn)
        if p!= self.X.shape[1]:
            raise(ValueError('New data must have seame number of columns as the ones the model has been trained with'))
        Xnc = scale_data(Xn,self.x_loc_,self.x_sca_)
        return(Xnc*self.x_Rweights_)



# -*- coding: utf-8 -*-
"""
Created on Tue Jun 30 13:17:46 2020

@author: Emmanuel Jordy Menvouta
"""

import unittest
import pandas as ps
import numpy as np
from ..sprm.sprm import sprm
from ..sprm.snipls import snipls
from ..sprm.rm import rm



class Testsprm(unittest.TestCase):
    """ Test some methods in the sprm class"""
    
    @classmethod
    def setUpClass(cls):
        print('setupClass')
        
    @classmethod
    def tearDownClass(cls):
        print('teardownClass')
        
        
    def setUp(self):
        self.data=ps.read_csv("./data/Returns_shares.csv")
        self.datav = np.matrix(self.data.values[:,2:8].astype('float64'))
        self.x = self.datav[:,0:5]
        self.y = self.datav[:,5]
        self.n=self.data.shape[0]
        self.p = self.data.shape[1]
        self.x0 = self.x.astype('float')
        self.y0 = self.y.astype('float')
        self.columns = self.data.columns[2:8]
        
        
        
        
        
        
        
    def tearDown(self):
        del self.x
        del self.y
        del self.n
        del self.p
        del self.x0
        del self.y0
        del self.data
        del self.datav
        
        
        
    
    def test_sprm(self):
        """ Test the functioning of the sprm object"""
        
        res_sprm = sprm(2,.8,'Hampel',.95,.975,.999,'kstepLTS','scaleTau2',True,100,.01,'ally','xonly',self.columns,True)
        res_sprm.fit(self.x0[:2666],self.y0[:2666])
        test_ans = 28.40453479240838
        np.testing.assert_almost_equal(np.linalg.norm(res_sprm.weightnewx(self.x0[2666:])),test_ans,decimal=4)
    
        
    
    
        
        
    def test_snipls(self):
        """ Test the functioning of the snipls object"""
        res_snipls = snipls(n_components=4, eta=.5)
        res_snipls.fit(self.x0[:2666],self.y0[:2666])
        test_ans = 38.6183244001568
        np.testing.assert_almost_equal(np.linalg.norm(res_snipls.predict(self.x0[2666:])),test_ans,decimal=4)
        
        
    def test_rm(self):
        """ Test the functioning of the rm object"""
        
        
        res_rm = rm('Hampel',.95,.975,.999,'median','mad','specific',True,100,.01,True)
        res_rm.fit(self.x0[:2666],self.y0[:2666])
        test_ans = 28.62510008113666
        np.testing.assert_almost_equal(np.linalg.norm(res_rm.predict(self.x0[2666:])),test_ans,decimal=4)
        
        
        
        

        
        
                
    
        
    

    
if __name__ =='__main__':
    unittest.main()
 



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Jul 9 14:20:17 2019

@author: Sven Serneels, Ponalytics
"""

__name__ = "ppdire"
__author__ = "Sven Serneels"
__license__ = "MIT"
__version__ = "0.2.9"
__date__ = "2021-04-26"









# -*- coding: utf-8 -*-
"""
Created on Sat Apr 11 17:00:13 2020

@author: Emmanuel Jordy Menvouta
"""


import numpy as np
from ..dicomo._dicomo_utils import *
import scipy.spatial as spp
import pandas as pd
#import Ball
import sympy
from ..utils.utils import MyException


# import Ball
# uncomment this line if you want to use     

def mdd_trim(beta, *args):
    """
    This is the objective function to be optimised when using MDD-SDR
    args is expected to be a tuple With the following elements : 
        X : input Data as matrix
        Y : input Data as 1d matrix 
        h, int: Dimension of the central subspace
        N2 : inverse of square root of the covariance matrix of X 
        is_distance_matrix, bool : if true, X and Y represent distance matrices
        trimming, float : trimming fraction to be applied for MDD
        center, bool :  if true, X and Y will be centered when computing MDD
        dmetric : 'the distance metric to be used in the computation of MDD
        biascorr, bool : if True, an unbiased estimator of MDD is computed.
    """
    X= args[0]
    Y= args[1]
    h=args[2]
    N2 = args[3]
    is_data_matrix = args[4]
    trimming = args[5]
    center = args[6]
    dmetric = args[7]
    biascorr = args[8]
    beta = np.reshape(beta,(-1,h),order = 'F')
    if is_data_matrix:
        X_coord = give_coords(X)
        #Y_coord = give_coords(Y)
        X_dat = np.matmul(X_coord,beta)
        res = np.sqrt(difference_divergence(X_dat,Y,center=center,trimming=trimming,biascorr=biascorr))
        return(-10*res)
    else :    
        X_dat = np.matmul(X,beta)
        res = np.sqrt(difference_divergence(X_dat,Y,center=center,trimming=trimming,biascorr=biascorr))
        return(-10*res)

    
def dcov_trim(beta, *args):
    """
    This is the objective function to be optimised when using DCOV-SDR
    args is expected to be a tuple With the following elements : 
        X : input Data as matrix
        Y : input Data as 1d matrix 
        h : Dimension of the central subspace
        N2 : inverse of square root of the covariance matrix of X 
        is_distance_matrix, bool : if true, X and Y represent distance matrices
        center, bool :  if true, X and Y will be centered when computing MDD
        dmetric : 'the distance metric to be used in the computation of MDD
        biascorr, bool : if True, an unbiased estimator of MDD is computed.
    """
    X= args[0]
    Y= args[1]
    h=args[2]
    N2 = args[3]
    is_distance_matrix = args[4]
    trimming = args[5]
    center = args[6]
    dmetric = args[7]
    beta = np.reshape(beta,(-1,h),order = 'F')
    if is_distance_matrix:
        X_coord = give_coords(X)
        Y_coord = give_coords(Y)
        X_dat = np.matmul(X_coord,beta)
        dmx, n1 = distance_matrix_centered(X_dat,trimming=trimming,
                                           center=center,dmetric=dmetric)
        dmy, n2 = distance_matrix_centered(Y_coord,trimming=trimming,
                                           center=center,dmetric=dmetric)
        res = distance_moment(dmx,dmy,n1=n1,center=center,trimming=trimming, order=2) 
        return(-10*res)
    else :
        Y =Y.reshape(-1,1)   
        X_dat = np.matmul(X,beta)
        dmx, n1 = distance_matrix_centered(X_dat,trimming=trimming,center=center,dmetric=dmetric)
        dmy, n2 = distance_matrix_centered(Y,trimming=trimming,center=center,dmetric=dmetric)
        res = distance_moment(dmx,dmy,n1=n1,center=center,trimming=trimming, order=2) 
        return(-10*res)

def give_coords(distances):
    """give coordinates of points for which distances given

    coordinates are given relatively. 1st point on origin, 2nd on x-axis, 3rd 
    x-y plane and so on. Maximum n-1 dimentions for which n is the number
    of points
     Args:
        distanes (list): is a n x n, 2d array where distances[i][j] gives the distance 
            from i to j assumed distances[i][j] == distances[j][i]

     Returns:
        numpy.ndarray: cordinates in list form n dim
    """
    distances = np.array(distances)

    n = len(distances)
    X = sympy.symarray('x', (n, n - 1))

    for row in range(n):
        X[row, row:] = [0] * (n - 1 - row)

    for point2 in range(1, n):

        expressions = []

        for point1 in range(point2):
            expression = np.sum((X[point1] - X[point2]) ** 2) 
            expression -= distances[point1,point2] ** 2
            expressions.append(expression)

        X[point2,:point2] = sympy.solve(expressions, list(X[point2,:point2]))[1]

    return X
    
    
def matpower(A, alpha):
    """
    computes the A to the power alpha using the eigen decomposition of A.
    """
    A= (A+ A.T)/2
    eig_vals, eig_vecs = np.linalg.eigh(A)
    res = np.matmul(np.matmul(eig_vecs,np.diag(eig_vals**alpha)),eig_vecs.T)
    return(res)
    

def discretize(y,h):    
    """
    This function is used to discretize a continuous variable y into h classes
    that are ordered.
    input :
        y : input data as vector or 1D matrix.
        h : the number of classes 
        
    output : 
         vector containing the class to which each element of y belongs. 
    """
    n = len(y)
    m = np.floor(n/h)
    yord = np.sort(y)
    divpt=[]
    for i in range(1,h):
        divpt.append(yord[int((i*m))])
    y1 = np.repeat(0,n)
    y1[y<divpt[0]] = 1
    y1[y>= divpt[h-2]] = h
    for i in range(1,h-1):
        y1[(y>=divpt[i-1]) & (y<divpt[i])] = i+1
    return(y1)
    

def SIR(x,y,n_slices,d,ytype='continuous', center_data=True, scale_data=True):
    """
    computes  the Sliced Inverse Regression  estimator of the central subspace.
    The algortihm is inspired from :
    "Bing, Li .(2018). Sufficient Dimension Reduction: Methods and Applications with R.
    Chapman & Hall/CRC Monographs on Statistics and Applied Probability."
    
    Input : 
        X : input Data as Matrix 
        Y : input data as vector or 1d matrix.
        n_slices, int : number of slices 
        d, int : dimension of the central subspace.
        ytype, str : either discrete or continuous
        center_data, bool, if true, the data is centered before SIR
        scale_data, bool, if true, the data is sclaed before SIR
        
    Output :
         The estimated basis of the central subspace.
        
    """
    y = np.asarray(y).flatten()
    n = x.shape[0]## zxception if d> p or x.shape[0] != y.shape[0]
    signsqrt = matpower(np.cov(x, rowvar=0 ), -0.5)
#     xc = x - x.mean(axis=0)
#     xstd= np.matmul(xc,signsqrt)

    if(center_data):
        x = x - x.mean(axis=0)
    if(scale_data):
        x= np.matmul(x,signsqrt)
    xstd = x
    if(ytype == 'continuous') : 
        ydis= discretize(y,n_slices)
    else:
        ydis = y
    yless = ydis 
    ylabel = []
    for i in range(n):
        if(np.var(yless) !=0):
            ylabel.append(yless[0])
            yless = yless[yless != yless[0]]
    ylabel.append(yless[0])
    prob=[]
    exy = []
    
    for i in range(n_slices):
        prob.append(len(ydis[ydis==ylabel[i]])/n)
        xres = np.apply_along_axis(np.mean,0,xstd[ydis ==ylabel[i],:])
        exy.append(xres)
        
        
    exy= np.vstack(exy)
    sirmat =np.matmul(np.matmul(exy.T,np.diag(np.array(prob))), exy)
    eig_vals, eig_vecs = np.linalg.eigh(sirmat)
    idx = eig_vals.argsort()[::-1]   
    eig_vals = eig_vals[idx]
    eig_vecs = eig_vecs[:,idx]
    if (scale_data) :
        return(np.matmul(signsqrt,eig_vecs[:, 0:d]))
    else :
        return(eig_vecs[:, 0:d])
        

def SAVE(x,y,n_slices,d,ytype='continuous',center_data=True, scale_data=True):
    """
    computes  the Sliced Average Variance Estimator of the central subspace.
    The algortihm is inspired from :
    "Bing, Li .(2018). Sufficient Dimension Reduction: Methods and Applications with R.
    Chapman & Hall/CRC Monographs on Statistics and Applied Probability."
    
    Input : 
        X : input Data as Matrix 
        Y : input data as vector or 1d matrix.
        n_slices, int : number of slices 
        d, int : dimension of the central subspace.
        ytype, str : either discrete or continuous
        center_data, bool, if true, the data is centered before SIR
        scale_data, bool, if true, the data is sclaed before SIR
        
    Output :
         The estimated basis of the central subspace.
        
    """
    y = np.asarray(y).flatten()
    p=x.shape[1]
    n = x.shape[0]## zxception if d> p or x.shape[0] != y.shape[0]
    signsqrt = matpower(np.cov(x, rowvar=0 ), -0.5)
#     xc = x - x.mean(axis=0)
#     xstd= np.matmul(xc,signsqrt)
    if(center_data):
        x = x - x.mean(axis=0)
    if(scale_data):
        x= np.matmul(x,signsqrt)
    
    xstd = x
    
    if(ytype== 'continuous') : 
        ydis= discretize(y,n_slices)
    else:
        ydis = y
    yless = ydis 
    ylabel = []
    for i in range(n):
        if(np.var(yless) !=0):
            ylabel.append(yless[0])
            yless = yless[yless != yless[0]]
    ylabel.append(yless[0])
    prob=[]   
    for i in range(n_slices):
        prob.append(len(ydis[ydis==ylabel[i]])/n)
        
    vxy=np.zeros((n_slices,p,p))
    for i in range(n_slices):
        vxy[i,:,:] = np.cov(xstd[ydis ==ylabel[i],:],rowvar=0)
    
    savemat = np.zeros((p,p))
    for i in range(n_slices):
        savemat = savemat + prob[i] * np.matmul((vxy[i,:,:]-np.identity(p)), (vxy[i,:,:]-np.identity(p)))
    
    eig_vals, eig_vecs = np.linalg.eigh(savemat)
    idx = eig_vals.argsort()[::-1]   
    eig_vals = eig_vals[idx]
    eig_vecs = eig_vecs[:,idx]
    if (scale_data) :
        return(np.matmul(signsqrt,eig_vecs[:, 0:d]))
    else :
        return(eig_vecs[:, 0:d])
        
        

def DR(x,y,n_slices,d,ytype='continuous', center_data=True, scale_data=True):
    """
    computes  the Directional Regression Estimator of the central subspace.
    The algortihm is inspired from :
    "Bing, Li .(2018). Sufficient Dimension Reduction: Methods and Applications with R.
    Chapman & Hall/CRC Monographs on Statistics and Applied Probability."
    
    Input : 
        X : input Data as Matrix 
        Y : input data as vector or 1d matrix.
        n_slices, int : number of slices 
        d, int : dimension of the central subspace.
        ytype, str : either discrete or continuous
        center_data, bool, if true, the data is centered before SIR
        scale_data, bool, if true, the data is sclaed before SIR
        
    Output :
         The estimated basis of the central subspace.
        
    """
    y = np.asarray(y).flatten()
    p=x.shape[1]
    n = x.shape[0]## zxception if d> p or x.shape[0] != y.shape[0]
    signsqrt = matpower(np.cov(x, rowvar=0 ), -0.5)
    
    if(center_data):
        x = x - x.mean(axis=0)
    if(scale_data):
        x= np.matmul(x,signsqrt)

    xstd = x 
    if(ytype== 'continuous') : 
        ydis= discretize(y,n_slices)
    else:
        ydis = y
    yless = ydis 
    ylabel = []
    for i in range(n):
        if(np.var(yless) !=0):
            ylabel.append(yless[0])
            yless = yless[yless != yless[0]]
    ylabel.append(yless[0])
    prob=[]
    exy = []
    
    for i in range(n_slices):
        prob.append(len(ydis[ydis==ylabel[i]])/n)
        
    vxy=np.zeros((n_slices,p,p,))
    exy=[]
    for i in range(n_slices):
        vxy[i,:,:,] = np.cov(xstd[ydis ==ylabel[i],:],rowvar=0)
        xres = np.apply_along_axis(np.mean,0,xstd[ydis ==ylabel[i],:])
        exy.append(xres)
    
    exy= np.vstack(exy)
    mat1 = np.zeros((p,p))
    mat2 = np.zeros((p,p))
    
    for i in range(n_slices):
        mat1 = mat1 + prob[i] *np.matmul(( vxy[i,:,:]+np.matmul( exy[i,:].reshape((-1,1)),exy[i,:].reshape((-1,1)).T)),(vxy[i,:,:]+ np.matmul( exy[i,:].reshape((-1,1)),exy[i,:].reshape((-1,1)).T)))
        mat2 = mat2 + prob[i] * (np.matmul(exy[i,:].reshape((-1,1)), exy[i,:].reshape((-1,1)).T))
        
    out = 2*mat1 + 2*np.matmul(mat2,mat2)+ 2*np.sum(np.diag(mat2))*mat2 - 2 *np.identity(p)
    eig_vals, eig_vecs = np.linalg.eigh(out)
    idx = eig_vals.argsort()[::-1]   
    eig_vals = eig_vals[idx]
    eig_vecs = eig_vecs[:,idx]
    if (scale_data) :
        return(np.matmul(signsqrt,eig_vecs[:, 0:d]))
    else :
        return(eig_vecs[:, 0:d])
    
   
def PHD(x, y,d, center_data=True, scale_data=True):
    
    """
    computes  the Principal Hessian Dimension estimator of the central subspace.
    The algortihm is inspired from :
    "Bing, Li .(2018). Sufficient Dimension Reduction: Methods and Applications with R.
    Chapman & Hall/CRC Monographs on Statistics and Applied Probability."
    
    Input : 
        X : input Data as Matrix 
        Y : input data as vector or 1d matrix.
        d, int : dimension of the central subspace.
        center_data, bool, if true, the data is centered before SIR
        scale_data, bool, if true, the data is sclaed before SIR
        
    Output :
         The estimated basis of the central subspace.
    
    """
    n = x.shape[0]
    if(len(y.shape)<2):
        y=y.reshape((-1,1))
    signsqrt = matpower(np.cov(x, rowvar=0 ), -0.5)
    if(center_data):
        x = x - x.mean(axis=0)
    if(scale_data):
        x= np.matmul(x,signsqrt)
    
    yc = y -np.mean(y, axis=0)
    out = np.matmul(np.multiply(x,yc).T,x)/n
    eig_vals, eig_vecs = np.linalg.eigh(out)
    idx = eig_vals.argsort()[::-1]   
    eig_vals = eig_vals[idx]
    eig_vecs = eig_vecs[:,idx]
    
    if (scale_data):
        return(np.matmul(signsqrt,eig_vecs[:, 0:d]))
    else : 
        return(eig_vecs[:, 0:d])
    

def IHT(x,y,d,center_data=True, scale_data=True):
    """
    computes  the Iterative Hessian Transformation estimator of the central subspace.
    The algortihm is inspired from :
    "Bing, Li .(2018). Sufficient Dimension Reduction: Methods and Applications with R.
    Chapman & Hall/CRC Monographs on Statistics and Applied Probability."
    
    Input : 
        X : input Data as Matrix 
        Y : input data as vector or 1d matrix.
        d, int : dimension of the central subspace.
        center_data, bool, if true, the data is centered before SIR
        scale_data, bool, if true, the data is sclaed before SIR
        
    Output :
         The estimated basis of the central subspace.
        
    """
    
    p=x.shape[1]
       ## zxception if d> p or x.shape[0] != y.shape[0]
    if(len(y.shape)<2):
        y=y.reshape((-1,1))
    signsqrt = matpower(np.cov(x, rowvar=0 ), -0.5)
    
    if(center_data):
        x = x - x.mean(axis=0)
    if(scale_data):
        x= np.matmul(x,signsqrt)
        
    covxy = np.cov(x,y,rowvar=0)[-1, :-1]
    covx =  np.cov(x,rowvar=0)
    mat = covxy
    for i in range(1,p):
        mat = np.column_stack((mat,np.matmul(matpower(covx,i),covxy)))
    
    out = np.matmul(mat,mat.T)
    eig_vals, eig_vecs = np.linalg.eigh(out)
    idx = eig_vals.argsort()[::-1]   
    eig_vals = eig_vals[idx]
    eig_vecs = eig_vecs[:,idx]
    if (scale_data):
        return(np.matmul(signsqrt,eig_vecs[:, 0:d]))
    else : 
        return(eig_vecs[:, 0:d])
        
# def ballcov_func(beta, *args):
#     """
#     Objective function  for BCOV-SDR.
#     Will only work after uncommenting the import Ball statement above
#     """
    
#     X= args[0]
#     Y= args[1]
#     h=args[2]
#     beta = np.reshape(beta,(-1,h),order = 'F')
#     X_dat = np.matmul(X, beta)
#     res = Ball.bcov_test(X_dat,Y,num_permutations=0)[0] 
#     return(-10*res)
    
    



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Created on Sun Dec  2 13:14:51 2018

# @author: sven

import numpy as np
import scipy.stats as sps
import scipy.spatial as spp
from sklearn.base import BaseEstimator
from sklearn.utils.metaestimators import _BaseComposition
from statsmodels import robust as srs
from ._dicomo_utils import *
import dcor as dc
# Optional: if Ballcov required, import Ball 
# import Ball

class MyException(Exception):
        pass

class dicomo(_BaseComposition,BaseEstimator):
    
    """
    The `dicomo` class implements (co)-moment statistics, covering both clasical product-moment 
    statistics, as well as more recently developed energy statistics. 
    The `dicomo` class also serves as a plug-in into `capi` and  `ppdire`. 
    It has been written consistently with `ppdire` such that it provides a wide range of 
    projection indices based on (co-)moments. Ancillary functions for (co-)moment 
    estimation are in `_dicomo_utils.py`.
    
    Parameters
    ------------ 

        est :  str
             mode of estimation. The set of options are `'arithmetic'` (product-moment)  or `'distance'` (energy statistics)

        mode : str
                 type of moment. Options include :
                * `'mom'`: moment 
                * `'var'`: variance 
                * `'std'`: standard deviation 
                * `'skew'`: skewness 
                * `'kurt'`: kurtosis
                * `'com'`: co-moment 
                * `'M3'`: shortcut for third order co-moment
                * `'cov'`: covariance 
                * `'cos'`: co-skewness
                * `'cok'`: co-kurtosis 
                * `'corr'`: correlation, 
                * `'continuum'`: continuum association 
                * `'mdd'`: martingale difference divergence (requires `est = 'distance'`)
                * `'mdc'`: martingale difference correlation (requires `est = 'distance'`)
                * `'ballcov'`: ball covariance (requires installing `Ball` and uncommenting the `import` statement)

        center : str 
                 internal centring used in calculation. Options are `mean` or `median`.  

    Attributes
    ------------
    Attributes always provided: 

        - `moment_`: The resulting (co-)moment Depending on the options picked, intermediate results are stored as well, such as `x_moment_`, `y_moment_` or `co_moment_`
    
    """
    

    def __init__(self,est='arithmetic',mode='mom',center='mean'):
        self.mode = mode
        self.est=est
        self.center=center
        self.liest = ['arithmetic','distance','sign','entropy']
        self.limo = ['mom','var','std','skew','kurt','com','cov','cok','cos','corr','continuum','M3', 'mdd','ballcov']
        self.licenter = ['mean','median']
        if not(self.mode in self.limo):
            raise(MyException("Only models allowed are: 'mom','var','skew','kurt','com','cov','cos','cok','corr','continuum','M3','mdd','mdc','ballcov'"))
        if not(self.est in self.liest):
            raise(MyException('Only estimator classes allowed are: "arithmetic", "distance", "sign", "entropy"'))
        if not(self.center in self.licenter):
            raise(MyException('Only centring classes allowed are: "mean", "median"'))
        if (self.est=='arithmetic' and self.mode in ['mdd','mdc']):
            raise(MyException('MDD only defined when est="distance"'))
        if (self.mode=='ballcov'): 
            print("To use the Ballcov functionality, uncomment the import statement and comment this line")
        
        
        
    def fit(self,x,**kwargs):
        """
        Fit a dicomo model

        Parameters
        ------------
            X : numpy array or pandas DataFrame
                input data


        Remarks:
        The `fit` function takes several optional input arguments. These are options that 
        apply to individual settings: 
            `biascorr`, Bool, when `True`, correct for bias. For classical product-moment statistics, this 
                is the small sample correction. For energy statistics, this leads to the estimates 
                that are unbiased in high dimension
                (but not preferred in low dimension). 
            `alpha`, float, parameter for continuum association. Has no effect for other options.  
            `option`, int, determines which higher order co-moment to calculate, 
                e.g. for co-skewness, `option=1` calculates CoS(x,x,y)
            `order`, int, which order (co-)moment to calculate. Can be overruled by `mode`, 
                e.g. if `mode='var'`, `order` is set to 2. 
            `calcmode`, str, to use the efficient or naive algorithm to calculate distance statistics. Defaults to `fast` when available.
        """
        
        
        if 'trimming' not in kwargs:
            trimming = 0
        else:
            trimming = kwargs.get('trimming')
            
        if 'biascorr' not in kwargs:
            biascorr = False
        else:
            biascorr = kwargs.get('biascorr')
            
        if 'alpha' not in kwargs:
            alpha = 1
        else:
            alpha = kwargs.get('alpha')
            
        if 'dmetric' not in kwargs:
            dmetric = 'euclidean'
        else:
            dmetric = kwargs.get('dmetric')
            
        if 'calcmode' not in kwargs:
            calcmode = 'fast'
        else:
            calcmode = kwargs.get('calcmode')
            
        if 'order' not in kwargs:
            order = 2
        else:
            order = kwargs.get('order')
            
        if self.mode == 'var':
            mode = 'mom' 
            order=2
        elif self.mode == 'cov':
            mode = 'com'
            order=2
        elif self.mode == 'std':
            self.center = 'mean'
            mode = 'mom'
            order=2
            num_power = 0.5
        elif self.mode == 'skew':
            self.center = 'mean'
            mode = 'mom'
            order = 3
            num_power = 1.5
        elif self.mode == 'kurt':
            self.center = 'mean'
            mode = 'mom'
            order = 4
            num_power = 2
            if 'Fisher' not in kwargs:
                Fisher = True
            else:
                Fisher = kwargs.get('Fisher')
        elif self.mode == 'cos':
            self.center = 'mean'
            mode = 'com'
            order = 3
            if 'standardized' not in kwargs:
                standardized = True
            else:
                standardized = kwargs.get('standardized')
        elif self.mode == 'M3':
            self.center = 'mean'
            mode = 'com'
            order = 3
            if 'standardized' not in kwargs:
                standardized = False
            else:
                standardized = kwargs.get('standardized')
        elif self.mode == 'cok':
            self.center = 'mean'
            mode = 'com'
            order = 4
            if 'standardized' not in kwargs:
                standardized = True
            else:
                standardized = kwargs.get('standardized')
            if 'Fisher' not in kwargs:
                Fisher = True
            else:
                Fisher = kwargs.get('Fisher')
        else:
            mode = self.mode
            
        if order > 2: 
            if 'option' not in kwargs:
                option = 1
            else:
                option = kwargs.get('option')
        else:
            option = 0
            
        n = len(x)
        ntrim = round(n * (1-trimming)) 
        
        if len(x.shape)==1:
            x = np.matrix(x).reshape((n,1))
        
        if mode=='corr':
            alpha = 1
        
        if n==0:
            raise(MyException('Please feed data with length > 0'))
            
        if self.center == 'median':
            locest = np.median
        else:
            locest = trim_mean
        
        # Classical variance, covariance and continuum as well as robust alternatives
        if self.est=='arithmetic':
            
            # Variance
            if mode!='com':
#                if self.center=='mean':
#                    xvar = trimvar(x,trimming)*ntrim/(ntrim-1)
#                elif self.center=='median':
#                    xvar = srs.mad(x)**2
                xmom = trim_mom(x,x,locest,order,trimming,option,biascorr) 
                self.x_moment_ = xmom
                
                if self.mode in ('std','skew','kurt'):
                    x2mom = trim_mom(x,x,locest,2,trimming,option,False)
                    xmom /= (x2mom**num_power) 
                    if biascorr:
                        if self.mode == 'skew':
                            xmom *= (ntrim-1)**2
                            xmom /= np.sqrt(ntrim**2 - ntrim)
                        elif self.mode == 'kurt':
                            xmom = xmom*ntrim - xmom/ntrim
                            xmom -= 3*(ntrim-1)**2.0 / ((ntrim-2)*(ntrim-3))
                            if not Fisher:
                                xmom += 3
                
                if mode=='mom':
                    self.moment_ = xmom
            
            # Covariance or continuum
            if mode!='mom':
                
                if 'y' not in kwargs:
                    raise(MyException('Please supply second data vector'))
                else:
                    y = kwargs.get('y')
                
                n1 = len(y)
                if n1==0:
                    raise(MyException('Please feed data with length > 0'))
                if n1!=n:
                    raise(MyException('Please feed x and y data of equal length'))
                if len(y.shape)==1:
                    y = np.matrix(y).reshape((n,1))
                
                como = trim_mom(x,y,locest,order,trimming,option,biascorr)
                
                self.co_moment_ = como
                
                if (biascorr and (order > 2)):
                    como *= ntrim
                
                if mode=='com':
                    self.moment_ = como
                
                    
                if self.mode in ('cok','cos'):
                    x2sd = np.sqrt(trim_mom(x,x,locest,2,trimming,option,biascorr))
                    y2sd = np.sqrt(trim_mom(y,y,locest,2,trimming,option,biascorr))
                        
                    if ((self.mode == 'cok') and biascorr): # biascorr is only exact if standardized
                        como -= como/(ntrim**2)
                        como -= 3*(ntrim-1)**2.0 / ((ntrim-2)*(ntrim-3))

                
                if mode in ['corr','continuum']:
                    
#                    if self.center=='mean':
#                        yvar = trimvar(y,trimming)*ntrim/(ntrim-1)
#                    
#                    elif self.center=='median':
#                        yvar = srs.mad(y)
                    ymom = trim_mom(y,y,locest,order,trimming,option,biascorr)
                    
                    self.y_moment_ = ymom
                    
        
        # Distance based metrics
        elif self.est=='distance':
            
            if 'dmetric' not in kwargs:
                dmetric = 'euclidean'
            else:
                dmetric = kwargs.get('dmetric')
                
            if (mode in ['ballcov','mdd','mdc']):
        
                if 'y' not in kwargs:
                        raise(MyException('Please supply second data vector'))
                else:
                    y = kwargs.get('y')
                    n1 = len(y)
                    if n1==0:
                        raise(MyException('Please feed data with length > 0'))
                    if n1!=n:
                        raise(MyException('Please feed x and y data of equal length')) 
                    if (mode in ['mdd','mdc']):
                        como=np.sqrt(difference_divergence(x,y,center=self.center,trimming=trimming,biascorr=biascorr))
                        self.co_moment_ = como
                        if mode=='mdd':
                            self.moment_ = como 
                        else:
                            xmom=difference_divergence(x,x,center=self.center,trimming=trimming,biascorr=biascorr)
                            ymom=difference_divergence(y,y,center=self.center,trimming=trimming,biascorr=biascorr)
                            self.x_moment_ = xmom
                            self.y_moment_ = ymom
                            mode = 'corr'

                    else:
                        dmy, n2 = distance_matrix_centered(y,biascorr=biascorr,
                                                   trimming=trimming,
                                                   center=self.center,
                                                   dmetric=dmetric) 
                        bcov_res=Ball.bcov_test(x,y,num_permutations=0)[0]
                        self.moment_ = bcov_res
                        
                
                
            
            elif (calcmode=='fast' and self.center =='mean' and trimming == 0 and order==2):
                if mode != 'com':
                    if biascorr:
                        xmom = np.sqrt(dc.u_distance_covariance_sqr(x,x))
                    else:
                        xmom = np.sqrt(dc.distance_covariance_sqr(x,x))
                    self.moment_ = xmom
                if mode !='mom':
                    if 'y' not in kwargs:
                        raise(MyException('Please supply second data vector'))
                    else:
                        y = kwargs.get('y')
                        n1 = len(y)
                    
                    if biascorr:
                        como = np.sqrt(dc.u_distance_covariance_sqr(x,y))
                    else:
                        como = np.sqrt(dc.distance_covariance_sqr(x,y))
                    if mode=='com':
                        self.co_moment_ = como
                        self.moment_ = como
                    elif mode in ['corr','continuum','cos','cok']:
                        if biascorr:
                            ymom = np.sqrt(dc.u_distance_covariance_sqr(y,y))
                        else:
                            ymom = np.sqrt(dc.distance_covariance_sqr(y,y))
                        self.y_moment_ = ymom
                        
            
                
            else:
                dmx, n1 = distance_matrix_centered(x,biascorr=biascorr,
                                                   trimming=trimming,
                                                   center=self.center,
                                                   dmetric=dmetric)
                
                # Variance
                if mode!='com':
                    
                    xmom = distance_moment(dmx,dmx,n1=n1,biascorr=biascorr,center=self.center,trimming=trimming,order=order,option=option)
                        
                    self.x_moment_ = xmom
                    
                    if self.mode in ('std','skew','kurt'):
                        x2mom =distance_moment(dmx,dmx,n1=n1,biascorr=biascorr,center=self.center,trimming=trimming,order=2,option=option)
                        xmom /= x2mom**num_power
                        
                    if mode=='mom':
                        self.moment_ = xmom
                        
                if mode!='mom':
                    
                    if 'y' not in kwargs:
                        raise(MyException('Please supply second data vector'))
                    else:
                        y = kwargs.get('y')
                        n1 = len(y)
                    if n1==0:
                        raise(MyException('Please feed data with length > 0'))
                    if n1!=n:
                        raise(MyException('Please feed x and y data of equal length'))
                    
                    dmy, n2 = distance_matrix_centered(y,biascorr=biascorr,
                                                   trimming=trimming,
                                                   center=self.center,
                                                   dmetric=dmetric)
                    
                    como = distance_moment(dmx,dmy,n1=n1,biascorr=biascorr,center=self.center,trimming=trimming,order=order,option=option)
                    
                    self.co_moment_ = como
                    
                    if mode=='com':
                        self.moment_ = como
                        
                    if self.mode in ('cok','cos'):
                        x2sd = distance_moment(dmx,dmx,n1=n1,biascorr=biascorr,center=self.center,trimming=trimming,order=2,option=option)
                        x2sd = np.sqrt(x2sd)
                        y2sd = distance_moment(dmy,dmy,n1=n1,biascorr=biascorr,center=self.center,trimming=trimming,order=2,option=option)
                        y2sd = np.sqrt(y2sd)
                    
                    if mode in ['corr','continuum','cok','cos']:
                        
                        ymom = distance_moment(dmy,dmy,n1=n,biascorr=biascorr,center=self.center,trimming=trimming,order=order,option=option)
                        self.y_moment_ = ymom
                    
        if mode == 'corr': 
            como /= (np.sqrt(xmom)*np.sqrt(ymom))
            self.moment_=como
        elif mode == 'continuum':
            como *= como * (np.sqrt(xmom)**(alpha -1))
            self.moment_=como
        
        if (self.mode in ('cok','cos') and standardized):
            iter_stop_2 = option 
            iter_stop_1 = order - option
            como /= np.power(x2sd,iter_stop_1)
            como /= np.power(y2sd,iter_stop_2)
            if ((self.mode == 'cok') and not Fisher): # Not very meaningful for co-moment
                como += 3
            self.moment_=como
        
                        
        
        if type(self.moment_)==np.ndarray:
            self.moment_ = self.moment_[0]
        return(self.moment_)
        
                    
                    
                



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Created on Wed Mar 25 09:01:53 2020

# @author: Sven Serneels, Ponalytics, Mar 2020. 


__all__ = ['GenSpatialSignPrePprocessor','gen_ss_pp','gen_ss_covmat']

from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.utils.validation import check_is_fitted
from .robcent import VersatileScaler, versatile_scale
from ._preproc_utilities import *
from ..utils.utils import _check_input
from ._gsspp_utils import *
from ._gsspp_utils import _norms, _gsspp

__all__ = ['GenSpatialSignPreProcessor', 'gen_ss_covmat', 'gen_ss_pp']
    
class GenSpatialSignPreProcessor(TransformerMixin,BaseEstimator):
    
    """
    GenSpatialSignPreProcessor Generalized Spatial Sign Pre-Processing as a scikit-learn compatible object
    that can be used in ML pipelines. 
    
    Parameters
    ---------- 
        center: str or function, 
            location estimator for centring.str options: 'mean', 'median', 'l1median', 'kstepLTS', 'None' 

        fun: str or function, 
            radial transformation function, str options: 'ss' (the non-generalized spatial sign, equivalent to sklearn's Normalizer), 'ball', 'shell', 'quad' (quadratic), 'winsor', or 'linear_redescending'
            Methods: sklearn API: `fit(X)`, `transform(X)` and `fit_transform(X)` with 

        
    Attributes
    ---------- 
    Attributes always provided : 

        -  `gss_` : the generalized spatial signs 
        -  `Xm_` : the centred data 
        -  `centring_` : VersatileScaler centring object 
        -  `X_gss_pp_` : Data preprocessed by Generalized Spatial Sign
    """    

    def __init__(self,center='l1median',fun='linear_redescending'):
        
        self.center = center
        self.fun = fun 
        
    def fit(self,X): 
        
        """ 
        Calculate and store generalized spatial signs
        """
        
        X = _check_input(X)
        n,p = X.shape
        if type(self.fun) is str: 
            fun = eval(self.fun)
        else:
            fun = self.fun
        vs = VersatileScaler(center=self.center,scale='None')
        Xm = vs.fit_transform(X)
        gss_ = fun(_norms(Xm),p,n)
        setattr(self,'gss_',gss_)
        setattr(self,'Xm_',Xm)
        setattr(self,'centring_',vs)
        
    def transform(self,X):
        
        """
        Calculate Generalized Spatial Sign Pre-Pprocessed Data
        """
        
        check_is_fitted(self,('gss_','Xm_'))
        Xgss = np.multiply(self.Xm_,self.gss_)
        setattr(self,'X_gsspp_',Xgss)
        return(Xgss)

    def fit_transform(self,X): 

        self.fit(X)
        self.transform(X)
        return(self.X_gsspp_)        
        
    
    

def gen_ss_pp(X,center='l1median',fun='linear_redescending'):
    
    """
    Generalized Spatial Sign Pre-Processing as a one pass function
    Inputs: 
        X: Data matrix 
        center: str or function, location estimator for centring.
                str options: 'mean', 'median', 'l1median', 'kstepLTS', 'None' 
        fun: str or function, radial transformation function,
                str options: 'ss' (the non-generalized spatial sign, equivalent
                to sklearn's Normalizer), 'ball', 'shell', 'quad' (quadratic), 
                'winsor', or 'linear_redescending'
    Outputs: the pre-processed data 
    """    
    
    if type(center) is str: 
        center = eval(center)
            
    if type(fun) is str: 
        fun = eval(fun)
    
    X = _check_input(X)        
    n = X.shape
    if len(n) > 1:
        p = n[1]
    else:
        p = 1
    n = n[0]
    
    if center != 'None':
        X = versatile_scale(X,center=center,scale='None')
        
    return(_gsspp(X,p,n,fun=fun))
    
    
def gen_ss_covmat(X,center='kstepLTS',fun=linear_redescending): 
    
    """
    Generalized Spatial Sign Covariance Matrix 
    Is equivalent to the covariance matrix of generalized spatial sign 
    pre-processed data. 
    
    First published in: 
        A generalized spatial sign covariance matrix,
        Jakob Raymaekers, Peter Rousseeuw, 
        Journal of Multivariate Analysis, 171 (2019), 94–111.
        
    Inputs: 
        X: Data matrix 
        center: str or function, location estimator for centring.
                str options: 'mean', 'median', 'l1median', 'kstepLTS', 'None' 
        fun: str or function, radial transformation function,
                str options: 'ss' (the non-generalized spatial sign, equivalent
                to sklearn's Normalizer), 'ball', 'shell', 'quad' (quadratic), 
                'winsor', or 'linear_redescending'
                
    Outputs: the generalized spatial sign covariance matrix
    """    
    
    X = _check_input(X)  
    rc = VersatileScaler(center=center, scale='None')
    n,p = X.shape
    Xm = rc.fit_transform(X)
    Xgss = _gsspp(Xm,p,n,fun=fun)
    return(Xgss.T*Xgss/n)
    
    

    
    
        


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Jan 25 18:22:27 2019

Functions called internally in M-estimation 

@author: Sven Serneels, Ponalytics
"""

import numpy as np
import pandas as ps

def Fair(x,probct,*args):
    return((1/(1 + abs(x/(probct * 2)))**2)) 
         
def Huber(x,probct,*args):
    x[np.where(x <= probct)[0]] = 1
    x[np.where(x > probct)] = probct/abs(x[np.where(x > probct)])
    return(x)
        
def Hampel(x,probct,hampelb,hampelr):
    wx = x
    wx[np.where(x <= probct)[0]] = 1
    wx[np.where((x > probct) & (x <= hampelb))[0]] = probct/abs(x[
            np.where((x > probct) & (x <= hampelb))[0]])
    wx[np.where((x > hampelb) & (x <= hampelr))[0]] = np.divide(
            probct * (hampelr - (x[np.where((x > hampelb) 
            & (x <= hampelr))[0]])),
            (hampelr - hampelb) * abs(x[np.where((x > hampelb) &
            (x <= hampelr))[0]])
        )
    wx[np.where(x > hampelr)[0]] = 0
    return(wx)
    
def brokenstick(n_components):
    q = np.triu(np.ones((n_components,n_components)))
    r = np.empty((n_components,1),float)
    r[0:n_components,0] = (range(1,n_components+1))
    q = np.matmul(q,1/r)
    q /= n_components
    return q




    


# -*- coding: utf-8 -*-
"""
"""

import numpy as np
from scipy.linalg import  inv, sqrtm
import copy
from sklearn.utils.metaestimators import _BaseComposition
from sklearn.base import RegressorMixin,BaseEstimator,TransformerMixin, defaultdict
from ..preprocessing.robcent import VersatileScaler
from ._sudire_utils import * 
from scipy.linalg import orth
import warnings
import statsmodels.robust.scale as srs
from scipy.stats import trim_mean
import dcor as dc
#import Ball
import statsmodels.api as sm
import inspect
from ..ipopt_temp.ipopt_wrapper import minimize_ipopt
from ..ipopt_temp.jacobian import FunctionWithApproxJacobianCentral,FunctionWithApproxJacobian
from ..dicomo._dicomo_utils import *
from ..utils.utils import *

class sudire(_BaseComposition,BaseEstimator,TransformerMixin,RegressorMixin):
    
    """SUDIRE Sufficient Dimension Reduction 
    
    The class allows for Sufficient Dimension Reduction using a variety of 
    methods. If the method requires optimization of a function, 
    This optimization is done through the Interior Point Optimizer (IPOPT)
    algorithm. 
        
    Parameters
    ----------
        sudiremeth: function or class. sudiremeth in this package can also be used, but user defined functions  can be processed. Built in options are : 

            save : Sliced Average Variance Estimation

            sir  : Slices Inverse Regression

            dr   : Directional Regression 

            iht  : Iterative Hessian Transformations

            dcov-sdr : SDR via Distance Covariance 
            
            mdd-sdr : SDR via Martingale Difference Divergence.

            bcov-sdr : SDR via ball covariance
            
        n_components : int 
                      dimension of the central subspace.
            
        trimming : float
                 trimming percentage to be entered as pct/100 
        
        optimizer_options : dict 
                            with options to pass on to the optimizer.Includes: 

        max_iter : int
                     Maximal number of iterations.

        tol: float
             relative convergence tolerance

        constr_viol_tol : float
                        Desired threshold for the constraint violation.
            
        optimizer_constraints : dict or list of dicts
                                 further constraints to be passed on to the optimizer function.
            
        optimizer_arguments: dict
                             extra arguments to be passed to the sudiremeth function during optimization.
        
        optimizer_start : numpy array
                         starting value for the optimization.
        
        center : str
             how to center the data. options accepted are options from sprm.preprocessing 
            
        center_data : bool 
                    If True, the data will be centered before the dimension reduction   
        
        scale_data : bool
                    if set to False, convergence to correct optimum  is not a given. Will throw a warning. 
            
        compression : bool
                     Use internal data compresion step for flat data. 
        
        n_slices :  int
                    The number of slices for SAVE, SIR, DR
    
        is_distance_mat : bool 
                         if the inputed matrices for x and y are distance matrices.
            
        dmetric :  str
                 distance metric used internally. Defaults to 'euclidean'
        
        fit_ols :  bool 
                 if True, an OLS model is fitted after the dimension reduction.
        
        copy : bool
              Whether to make a deep copy of the input data or not. 
        
        verbose : bool
                 Set to True prints the iteration number. 
        
        return_scaling_object: bool. 
                                If True, the scaling object will be return after the dimension reduction. 

    Attributes
    ----------
    Attributes always provided 
        -  `x_loadings_`: Estimated basis of the central subspace 
        -  `x_scores_`: The projected X data. 
        -  `x_loc_`:  location estimate for X 
        -  `x_sca_`:  scale estimate for X
        - ` ols_obj` : fitted OLS objected
        -  `y_loc_`: y location estimate
        -  `y_sca_`: y scale estimate

    Attributes created only when corresponding input flags are `True`:
        -   `whitening_`: whitened data matrix (usually denoted K)
        -   `scaling_object_`: scaling object from `VersatileScaler`  
        
    
        
    
    """

    def __init__(self,
                 sudiremeth = 'dcov-sdr', 
                 n_components = 2, 
                 trimming = 0,
                 optimizer_options = {'max_iter': 1000}, 
                 optimizer_constraints = None,
                 optimizer_arguments =None,
                 optimizer_start = None,
                 center_data=True,
                 center='mean',
                 scale_data=True,
                 whiten_data=False,
                 compression = False,
                 n_slices = 6,
                 dmetric = 'euclidean',
                 fit_ols = True,
                 copy=True,
                 verbose=True, 
                 return_scaling_object=True):
        # Called arguments
        self.sudiremeth = sudiremeth
        self.n_components = n_components
        self.trimming = trimming
        self.optimizer_options = optimizer_options
        self.optimizer_constraints = optimizer_constraints
        self.optimizer_arguments = optimizer_arguments
        self.optimizer_start = optimizer_start
        self.center = center
        self.center_data = center_data
        self.scale_data = scale_data
        self.whiten_data = whiten_data
        self.compression = compression
        self.n_slices = n_slices
        self.dmetric = dmetric
        self.fit_ols = fit_ols
        self.copy = copy
        self.verbose = verbose
        self.return_scaling_object = return_scaling_object
        
        # Other global parameters 
        self.licenter = ['mean','median']
        if not(self.center in self.licenter):
            raise(ValueError('Only location estimator classes allowed are: "mean", "median"'))
            
        self.limeths =['sir', 'save', 'dr','dcov-sdr','bcov-sdr','mdd-sdr','phd','iht']
        if (not(self.sudiremeth in self.limeths) and not callable(self.sudiremeth)):            
            raise(ValueError('Only SDR methods allowed are : "sir", "save", "dr", "dcov-sdr","bcov-sdr", "mdd-sdr", "iht","phd"'))
            
    

    def fit(self,X,y,*args,**kwargs):
        
        """
        Fit a Sufficient Dimension Reduction Model. 
        
        Parameters
        ----------
                X : matrix or data frame 
                    Input data of predictors  

                y :  vector or 1D matrix
                    Response data
                
                args or kwargs : 
                                Further parameters to user defined sudiremeth can be passed here  
        Returns
        -------
            self
        
        -------
        """

        # Collect optional fit arguments
            
        if 'dmetric' not in kwargs:
            dmetric = 'euclidean'
        else:
            dmetric = kwargs.get('dmetric')
            
        if 'biascorr' not in kwargs:
            biascorr = False
        else:
            biascorr = kwargs.get('biascorr')
            
        if 'flag' not in kwargs:
            flag = 'two-block'
        else :
            flag = kwargs.get('flag')     
        
        if 'is_distance_mat' not in kwargs : 
            is_distance_mat =False
        else : 
            is_distance_mat = kwargs.pop('is_distance_mat')
            
            
                               
        
        # Initiate some parameters and data frames
        if self.copy:
            X0 = copy.deepcopy(X)
            self.X0 = X0
        else:
            X0 = X        
        X = convert_X_input(X0)    
        n,p = X0.shape 
        trimming = self.trimming
        
        # Check dimensions 
        if self.n_components > min(n,p):
            raise(MyException('number of components cannot exceed number of variables or sample size'))
            
        # Pre-processing adjustment if whitening
        if self.whiten_data:
            self.center_data = True
            self.scale_data = False
            self.compression = False
            print('All results produced are for whitened data')
            
         #Store original X data mean and varcov matrix.
        varmatx = np.cov(X,rowvar=0)
        meanx = X.mean(axis=0)
        N2 = inv(sqrtm(varmatx))
        
                    
         # Data Compression for flat tables if required                
        if ((p>n) and self.compression):
            V,S,U = np.linalg.svd(X.T,full_matrices=False)
            X = np.matmul(U.T,np.diag(S))
            n,p = X.shape
            
            if (srs.mad(X)==0).any(): 
                warnings.warn('Due to low scales in data, compression would induce zero scales.' 
                              + '\n' + 'Proceeding without compression.')
                dimensions = False
                if copy:
                    X = copy.deepcopy(X0)
                else:
                    X = X0
            else:
                dimensions = True
        else:
            dimensions = False
            
            # Centring and scaling
            # centering :
        if self.center_data:
            if self.center != 'mean':
                centring = VersatileScaler(center=self.center, scale ='None',trimming=self.trimming) 
                Xs = centring.fit_transform(X0)
                mX = centring.col_loc_
                sX = centring.col_sca_
            else :
                Xs = X -trim_mean(X, self.trimming,axis=0) 
                mX = trim_mean(X, self.trimming,axis=0)
                sX = np.sqrt(np.diag(varmatx))                
        else:
            Xs = X0
            mX = np.zeros((1,p))
            sX = np.ones((1,p))
          
        if self.scale_data:
            if self.center=='mean': 
                scale = 'std'
                N2 = inv(sqrtm(varmatx))
                Xs = np.matmul(Xs,N2) 
                
            elif((self.center=='median') or (self.center=='l1median')):
                scale = 'mad' 
                centring = VersatileScaler(center=self.center,scale=scale,trimming=trimming) 
                Xs = centring.fit_transform(X0)
                mX = centring.col_loc_
                sX = centring.col_sca_
            else :
                raise(MyException('centering options have to be either "mean", "median", or "l1median"'))
                
        else:
            scale = 'None'
            warnings.warn('Without scaling, convergence to optima is not given')
           
        
        # Initiate centring object and scale X data      
            
        if self.whiten_data:
            V,S,U = np.linalg.svd(Xs.T,full_matrices=False)
            del U
            K = (V/S)[:,:p]
            del S
            Xs = np.matmul(Xs, K)
            Xs *= np.sqrt(p)

        # Pre-process y data  
        ny = y.shape[0]
        y = convert_y_input(y)
        if len(y.shape) < 2:
            y = np.matrix(y).reshape((ny,1))
        if ny != n:
            raise(MyException('X and y number of rows must agree'))
        
        # Pre-process y data when available 
        if flag != 'one-block':    
            ny = y.shape[0]
            y = convert_y_input(y)
            if len(y.shape) < 2:
                y = np.matrix(y).reshape((ny,1))
            if ny != n:
                raise(MyException('X and y number of rows must agree'))
            if self.copy:
                y0 = copy.deepcopy(y)
                self.y0 = y0
                
            if self.center_data:
                ys=y # dont center y
                my = np.mean(y, axis=0)
                sy = np.sqrt(np.var(y,axis=0))
            else:
                ys = y
                my = 0
                sy = 1
            ys = ys.astype('float64')
        
        else:
            ys = None
        
                   
        if(self.sudiremeth == 'sir'):
            P = SIR(Xs, ys, self.n_slices,self.n_components,self.center_data, self.scale_data)
            if self.scale_data:
                P = np.matmul(N2,P)
            projMat = np.matmul(np.matmul(P,inv(np.matmul(P.T,P))),P.T)
            T = np.matmul(self.X0, P)
            
        elif(self.sudiremeth =='save'):
            P = SAVE(Xs, ys, self.n_slices,self.n_components,self.center_data, self.scale_data)
            if self.scale_data:
                P = np.matmul(N2,P)
            projMat = np.matmul(np.matmul(P,inv(np.matmul(P.T,P))),P.T)
            T = np.matmul(self.X0, P)
            
        elif(self.sudiremeth == 'dr') : 
            P = DR(Xs, ys, self.n_slices,self.n_components,self.center_data, self.scale_data)
            if self.scale_data:
                P = np.matmul(N2,P)
            projMat = np.matmul(np.matmul(P,inv(np.matmul(P.T,P))),P.T)
            T = np.matmul(self.X0, P)
            
        elif(self.sudiremeth == 'phd'):
            P = PHD(Xs, ys,self.n_components,self.center_data, self.scale_data)
            if self.scale_data:
                P = np.matmul(N2,P)
            projMat = np.matmul(np.matmul(P,inv(np.matmul(P.T,P))),P.T)
            T = np.matmul(self.X0, P)
            
        elif(self.sudiremeth == 'iht'):
            P = IHT(Xs, ys,self.n_components,self.center_data, self.scale_data)
            if self.scale_data:
                P = np.matmul(N2,P)
            projMat = np.matmul(np.matmul(P,inv(np.matmul(P.T,P))),P.T)
            T = np.matmul(self.X0, P)
            
        
        else : #SDR obtained through optimization of some function
            
             ## choose starting value for DCOV-SDR
            if self.optimizer_start is None :
                save_start=SAVE(Xs,ys,3,self.n_components,self.center_data, self.scale_data)
                if self.scale_data:
                    save_start = np.matmul(N2,save_start)
                sir_start=SIR(Xs, ys,6,self.n_components,self.center_data, self.scale_data)
                if self.scale_data:
                    sir_start = np.matmul(N2,sir_start)
                beta_save= orth(save_start)
                dc_save = dc.distance_covariance_sqr(np.matmul(Xs,beta_save),y)       
                beta_sir = orth(sir_start)
                dc_sir = dc.distance_covariance_sqr(np.matmul(Xs,beta_sir),y)
                DR_start = DR(Xs,y,6,self.n_components,self.center_data, self.scale_data)
                if self.scale_data:
                    DR_start = np.matmul(N2, DR_start)
                beta_DR = orth(DR_start)
                dc_DR = dc.distance_covariance_sqr(np.matmul(Xs,beta_DR),y)
                if(dc_save >= dc_sir and dc_save >= dc_DR) : 
                    self.optimizer_start= save_start.flatten(order='F')
                elif (dc_sir >= dc_save and dc_sir >= dc_DR) :    
                    self.optimizer_start = sir_start.flatten(order='F')
                else :
                    self.optimizer_start = DR_start.flatten(order='F')
                
            ## add constraints for the optimization
            if  self.optimizer_constraints is None:
                const_x=[]
                const_z=[]
                for i in range(0, self.n_components):
                    for j in range(0,self.n_components): 
                        const_x.append({'type': 'eq', 'fun' : const_xscale,
                                      'args':(Xs,self.n_components,i,j)}) # change const2 to const_func
                        const_z.append({'type': 'eq', 'fun' : const_zscale,
                                      'args':(Xs,self.n_components,i,j)})
                
                if self.scale_data:
                    self.optimizer_constraints = tuple(const_z)
                else :
                    self.optimizer_constraints = tuple(const_x)
                    
            
            if  self.optimizer_arguments is None:
                N2 = inv(sqrtm(varmatx))
                self.optimizer_arguments = (Xs,ys, self.n_components,N2, is_distance_mat,
                                            self.trimming, self.center, dmetric,biascorr)
            
            #perform DCOV-SDR optimization
            res = minimize_ipopt(dcov_trim, self.optimizer_start,
                                 args = self.optimizer_arguments,
                                 constraints = self.optimizer_constraints,
                                 options=self.optimizer_options)
            if self.scale_data : 
                dcov_res = np.matmul(N2,np.reshape(res.x,(-1,self.n_components),order = 'F'))
            else : 
                dcov_res = np.reshape(res.x,(-1,self.n_components),order = 'F')
            
            if(self.sudiremeth == 'dcov-sdr'):
                P = dcov_res
                projMat = np.matmul(np.matmul(P,inv(np.matmul(P.T,P))),P.T)
                T = np.matmul(self.X0, P)
                
            elif(self.sudiremeth == 'bcov-sdr'):
                self.optimizer_start = dcov_res.flatten(order='F')
                res_ball = minimize_ipopt(ballcov_func,self.optimizer_start,
                                          args = self.optimizer_arguments,
                                          constraints = self.optimizer_constraints,
                                          options=self.optimizer_options)
                if self.scale_data:
                    P = np.matmul(N2,np.reshape(res_ball.x,(-1,self.n_components),order = 'F'))
                else :
                    P = np.reshape(res_ball.x,(-1,self.n_components),order = 'F')
                projMat = np.matmul(np.matmul(P,inv(np.matmul(P.T,P))),P.T)
                T = np.matmul(self.X0, P)
                
                
            elif(self.sudiremeth == 'mdd-sdr'):
                self.optimizer_start = dcov_res.flatten(order='F')
                res_mdd = minimize_ipopt(mdd_trim,self.optimizer_start,
                                          args = self.optimizer_arguments,
                                          constraints = self.optimizer_constraints,
                                          options=self.optimizer_options)
                if self.scale_data :
                    P = np.matmul(N2,np.reshape(res_mdd.x,(-1,self.n_components),order = 'F'))
                else :
                    P = np.reshape(res_mdd.x,(-1,self.n_components),order = 'F')
                projMat = np.matmul(np.matmul(P,inv(np.matmul(P.T,P))),P.T)
                T = np.matmul(self.X0, P)
                
            else : ## user defined function   
                self.optimizer_start = dcov_res.flatten(order='F')
                opt_res = minimize_ipopt(self.sudiremeth, self.optimizer_start,
                                          args = self.optimizer_arguments,
                                          constraints = self.optimizer_constraints,
                                          options=self.optimizer_options)
                if self.scale_data : 
                    P = np.matmul(N2,np.reshape(opt_res.x,(-1,self.n_components),order = 'F'))
                else : 
                    P = np.reshape(opt_res.x,(-1,self.n_components),order = 'F')
                projMat = np.matmul(np.matmul(P,inv(np.matmul(P.T,P))),P.T)
                T = np.matmul(self.X0, P)
                
        # perform OLS regression
        T_reg = sm.add_constant(T) # adding a constant
        ols_obj= sm.OLS(ys, T_reg).fit()
        
                                
        # Re-adjust estimates to original dimensions if data have been compressed 
        if dimensions:      
            P = np.matmul(V[:,0:p],P)
        
        

        setattr(self,"x_loadings_",P)
        setattr(self,"x_scores_",T)
        setattr(self, "proj_mat_", projMat)
        if self.whiten_data:
            setattr(self,"whitening_",K)   
        setattr(self,"x_loc_",mX)
        setattr(self,"x_sca_",sX)
        setattr(self,'scaling_',scale)
        setattr(self,'ols_obj_',ols_obj)
        if self.return_scaling_object and self.center!= 'mean':
            setattr(self,'scaling_object_',centring)
        
        return(self)
    
    def transform(self,Xn, distance_mat=False):
        """
        Computes the dimension reduction of the data Xn based on the fitted sudire model.

        Parameters
        ----------
                Xn : matrix or data frame
                     Input data to be transformed 

                distance_mat : numpy array 
                                 distance matrix to represent similarity between observations. 
                
                args or kwargs: 
                                Further parameters to user defined sufdiremeth can be passed here  
        Returns
        -------
        transformed_data : numpy array
                             the dimension reduced data 

         -------
        """
        Xn = convert_X_input(Xn)
        (n,p) = Xn.shape
        (q,h) = self.x_loadings_.shape
        if p!=q:
            raise(ValueError('New data must have same number of columns as the ones the model has been trained with'))
        return(np.matmul(Xn, self.x_loadings_))

    
    def predict(self,Xn, is_distance_mat=False):
        """
        predicts the response  on new data Xn

        Parameters
        ----------
                Xn : matrix or data frame
                     Input data to be transformed 

                is_distance_mat : bool 
                                 if True, Xn is treated as a distance matrix 
                
        Returns
        -------
        predictions : numpy array 
                      The predictions from the dimension reduction model
        -------
        """
        Xn = convert_X_input(Xn)
        (n,p) = Xn.shape
        (q,h) = self.x_loadings_.shape
        if p!=q:
            raise(ValueError('New data must have same number of columns as the ones the model has been trained with'))
        Xns = self.transform(Xn)
        Xns = sm.add_constant(Xns) # adding a constant
        ys = self.ols_obj_.predict(Xns)
        return(ys)
        
   
    @classmethod   
    def _get_param_names(cls):
        """Get parameter names for the estimator"""
        # fetch the constructor or the original constructor before
        # deprecation wrapping if any
        init = getattr(cls.__init__, 'deprecated_original', cls.__init__)
        if init is object.__init__:
            # No explicit constructor to introspect
            return []

        # introspect the constructor arguments to find the model parameters
        # to represent
        init_signature = inspect.signature(init)
        # Consider the constructor parameters excluding 'self'
        parameters = [p for p in init_signature.parameters.values()
                    if p.name != 'self' and p.kind != p.VAR_KEYWORD]
        for p in parameters:
            if p.kind == p.VAR_POSITIONAL:
                raise RuntimeError("scikit-learn estimators should always "
                                "specify their parameters in the signature"
                                " of their __init__ (no varargs)."
                                " %s with constructor %s doesn't "
                                " follow this convention."
                                % (cls, init_signature))
        # Extract and sort argument names excluding 'self'
        return sorted([p.name for p in parameters])

    def get_params(self, deep=False):
        """Get parameters for this estimator.
        Parameters
        ----------
        deep : boolean, optional
            If True, will return the parameters for this estimator and
            contained subobjects that are estimators.
        Returns
        -------
        params : mapping of string to any
            Parameter names mapped to their values.
        ------
        Copied from ScikitLlearn instead of imported to avoid 'deep=True'
        """
        out = dict()
        for key in self._get_param_names():
            value = getattr(self, key, None)
            if deep and hasattr(value, 'get_params'):
                deep_items = value.get_params().items()
                out.update((key + '__' + k, val) for k, val in deep_items)
            out[key] = value
        return out

    def set_params(self, **params):
        """Set the parameters of this estimator.
        Copied from ScikitLearn, adapted to avoid calling 'deep=True'
        
        Returns
        -------
        self
        ------
        Copied from ScikitLlearn instead of imported to avoid 'deep=True'
        """
        if not params:
            # Simple optimization to gain speed (inspect is slow)
            return self
        valid_params = self.get_params()

        nested_params = defaultdict(dict)  # grouped by prefix
        for key, value in params.items():
            key, delim, sub_key = key.partition('__')
            if key not in valid_params:
                raise ValueError('Invalid parameter %s for estimator %s. '
                                'Check the list of available parameters '
                                'with `estimator.get_params().keys()`.' %
                                (key, self))

            if delim:
                nested_params[key][sub_key] = value
            else:
                setattr(self, key, value)
                valid_params[key] = value

        for key, sub_params in nested_params.items():
            valid_params[key].set_params(**sub_params)

        return self

def estimate_structural_dim(sudiremeth, Xn, y, B, *args,**kwargs):
    """
    Estimates the dimension of the central subspace using 
    the sudiremeth.  This approach is based on the bootstrap method of Sheng and Yin (2016)
    
    Parameters
    ----------

        sudiremeth : str
                    the SDR method to use in the estimation.

        X :  numpy array or dataframe
            Input X data

        Y : vector or 1d matrix
            Input Y data as 

        B : int 
            Number of bootstrap replications

        kwargs:

            n_slices: number of slices for SIR/SAVE/DR

            center_data, bool : if the data  should be centered 

            scale_data, bool :  if the data should be scaled

            center, string :   which centering('mean', 'median')
            
    Returns
    ----------     

    h : int 
        representing the dimension of the central subspace

    ----------
    """
    if 'n_slices' not in kwargs:
        n_slices = 6
    else:
        n_slices = kwargs.pop('n_slices')
    
    if 'center_data' not in kwargs :
        center_data = True
    else :
        center_data = kwargs.pop('center_data')
        
    if 'scale_data' not in kwargs :
        scale_data = True
    else :
        scale_data = kwargs.pop('scale_data')
        
    if 'center' not in kwargs :
        center = 'mean'
    else :
        center = kwargs.pop('center')
        
            
    Xn = convert_X_input(Xn)
    y  = convert_y_input(y)  
    n,p = Xn.shape
    
    diff_b= []
    mean_diff = []
    for k in range(1,p+1):
        print('possible dim', k)
        sdr = sudire(sudiremeth, center_data= center_data,
                     scale_data=scale_data, center=center,
                     n_slices=n_slices, n_components = k)
        sdr.fit(Xn, y=y)
        projMat = sdr.proj_mat_
        
        for b in range(B) :  
            idx = np.random.randint(0,n,n)
            X_b = Xn[idx, :].copy()
            sdr_b = sudire(sudiremeth, center_data= center_data,
                           scale_data=scale_data, center=center,
                          n_slices = n_slices, n_components = k)
            sdr_b.fit(X_b,y=y)
            projMat_b = sdr_b.proj_mat_
            uh,sh,vh=np.linalg.svd(projMat-projMat_b)
            diff_b.append( np.nanmax(sh))
            
        mean_diff.append(np.mean(diff_b))   
    return (np.argmin(mean_diff)+1, mean_diff)

 



# -*- coding: utf-8 -*-
"""
Created on Tue Jun 30 13:17:46 2020

@author: Emmanuel Jordy Menvouta
"""

import unittest
import pandas as ps
import numpy as np
from ..preprocessing.robcent import VersatileScaler
import sklearn.decomposition as skd
from ..dicomo.dicomo import dicomo
from ..ppdire.ppdire import ppdire
import sklearn.cross_decomposition as skc

class Testppdire(unittest.TestCase):
    """ Test some methods in the ppdire class"""
    
    @classmethod
    def setUpClass(self):
        print('setupClass')
        
    @classmethod
    def tearDownClass(self):
        print('teardownClass')
        
        
    def setUp(self):
        self.data=ps.read_csv("./data/Returns_shares.csv")
        self.datav = np.matrix(self.data.values[:,2:8].astype('float64'))
        self.x = self.datav[:,1:5]
        self.y = self.datav[:,0]
        self.n=self.data.shape[0]
        self.p = self.data.shape[1]
        self.centring = VersatileScaler()
        self.Xs = self.centring.fit_transform(self.x)
        
        
        
        
        
        
        
    def tearDown(self):
        del self.x
        del self.y
        del self.n
        del self.p
        del self.Xs
        del self.centring
        
        
    
    def test_pca(self):
        """ tests the exactness of ppdire's pca"""
        
        pppca = ppdire(projection_index = dicomo, pi_arguments = {'mode' : 'var'}, n_components=4, optimizer='SLSQP')
        pppca.fit(self.x)
        skpca = skd.PCA(n_components=4)
        skpca.fit(self.Xs)
        np.testing.assert_almost_equal(np.abs(pppca.x_loadings_),np.abs(skpca.components_.T),decimal=3)
        
    def test_pls(self):
        """ tests the exactness of ppdire's pls"""
        
        skpls = skc.PLSRegression(n_components=4)
        skpls.fit(self.Xs,(self.y-np.mean(self.y))/np.std(self.y))
        pppls = ppdire(projection_index = dicomo, pi_arguments = {'mode' : 'cov'}, n_components=4, square_pi=True, optimizer='SLSQP', optimizer_options={'maxiter':500})
        pppls.fit(self.x,self.y)
        np.testing.assert_almost_equal(np.abs(np.matmul(self.Xs,skpls.coef_)*np.std(self.y) + np.mean(self.y)),np.abs(pppls.fitted_),decimal=3)
        
#    def test_robust(self):
#        lcpca = ppdire(projection_index = dicomo, pi_arguments = {'mode' : 'var', 'center': 'median'}, n_components=4, optimizer='grid',optimizer_options={'ndir':1000,'maxiter':10})
#        lcpca.fit(self.x)
#        test_ans=np.array([[ 0.6324543 , -0.00651997, -0.35820225,  0.6438448 ],
#                           [ 0.44750274, -0.67228343,  0.4950862 , -0.21806968],
#                           [ 0.53378114,  0.28794634, -0.46650197, -0.72699245],
#                           [ 0.35432068,  0.68524337,  0.64350842,  0.09692107]])
#        np.testing.assert_almost_equal(np.abs(test_ans),np.abs(lcpca.x_loadings_),decimal=3)
        
    

    
if __name__ =='__main__':
    unittest.main()
 




"""
Author: Matthias Kuemmerer, 2014
"""
from __future__ import print_function, division, unicode_literals, absolute_import

import sys
import numpy as np


class FunctionWithApproxJacobian(object):
    def __init__(self, func, epsilon, verbose=True):
        self._func = func
        self.epsilon = epsilon
        self.value_cache = {}
        self.verbose = verbose

    def __call__(self, x, *args, **kwargs):
        key = tuple(x)
        if not key in self.value_cache:
            self.log('.')
            value = self._func(x, *args, **kwargs)
            if np.any(np.isnan(value)):
                print("Warning! nan function value encountered at {0}".format(x))
            self.value_cache[key] = value
        return self.value_cache[key]

    def func(self, x, *args, **kwargs):
        if self.verbose:
            print(x)
        return self(x, *args, **kwargs)

    def log(self, msg):
        if self.verbose:
            sys.stdout.write(msg)
            sys.stdout.flush()

    def jac(self, x, *args, **kwargs):
        self.log('G[')
        x0 = np.asfarray(x)
        #print x0
        dxs = np.zeros((len(x0), len(x0) + 1))
        for i in range(len(x0)):
            dxs[i, i + 1] = self.epsilon
        results = [self(*(x0 + dxs[:, i], ) + args, **kwargs) for i in range(len(x0) + 1)]
        jac = np.zeros([len(x0), len(np.atleast_1d(results[0]))])
        for i in range(len(x0)):
            jac[i] = (results[i + 1] - results[0]) / self.epsilon
        self.log(']')
        return jac.transpose()


class FunctionWithApproxJacobianCentral(FunctionWithApproxJacobian):
    def jac(self, x, *args, **kwargs):
        self.log('G[')
        x0 = np.asfarray(x)
        #print x0
        dxs = np.zeros((len(x0), 2*len(x0)))
        for i in range(len(x0)):
            dxs[i, i] = -self.epsilon
            dxs[i, len(x0)+i] = self.epsilon
        results = [self(*(x0 + dxs[:, i], ) + args, **kwargs) for i in range(2*len(x0))]
        jac = np.zeros([len(x0), len(np.atleast_1d(results[0]))])
        for i in range(len(x0)):
            jac[i] = (results[len(x0)+i] - results[i]) / (2*self.epsilon)
        self.log(']')
        return jac.transpose()


# -*- coding: utf-8 -*-
"""
Created on Tue Jun 30 13:17:46 2020

@author: Emmanuel Jordy Menvouta
"""

import unittest
import numpy as np
import pandas as pd
from direpack.sudire._sudire_utils import *
from sklearn.model_selection import train_test_split

class Testsudire(unittest.TestCase):
    """ Test some methods in the sudire class"""
    
    @classmethod
    def setUpClass(cls):
        print('setupClass')
        
    @classmethod
    def tearDownClass(cls):
        print('teardownClass')
        
        
    def setUp(self):
        self.data=pd.read_csv('./data/auto-mpg.csv', index_col='car_name')
        self.data = self.data[self.data.horsepower != '?']
        self.data.horsepower = self.data.horsepower.astype('float')
        self.x = self.data
        self.y = self.x['mpg']
        self.x.drop('mpg', axis=1, inplace=True)
        self.x.drop('origin', axis = 1, inplace = True)            
        self.n=self.x.shape[0]
        self.p = self.x.shape[1]
        self.struct_dim = 2
        self.x_train, self.x_test, self.y_train, self.y_test = train_test_split(
        self.x, self.y, test_size=0.3, random_state=42)
        
        
        
        
        
        
        
        
    def tearDown(self):
        del self.x
        del self.y
        del self.n
        del self.p
        del self.data
        del self.struct_dim
        del self.x_train
        del self.x_test
        del self.y_train
        del self.y_test
        
#    def test_estimdim(self):
#        """ Tests the estimation of the central subspace via Bootstrap """
#        
#        central_dim, diff_vec = estimate_structural_dim('dr',self.x_train.values,self.y_train.values , B=100, n_slices=4)
#        np.testing.assert_equal(central_dim,6)
        
    def test_sir(self):
        """ Tests Sliced Inverse Regression"""
        
        #mod_auto = sudire('sir', center_data= True, scale_data=True,n_components=self.struct_dim)
        #mod_auto.fit(self.x_train.values, self.y_train.values)
        res_sir = SIR(self.x_train.values, self.y_train.values,6,self.struct_dim,'continuous',False,False)
        test_ans = 1.4142135623730947
        np.testing.assert_almost_equal(np.linalg.norm(res_sir),test_ans,decimal=14)
        
    def test_save(self):
        """ Tests Sliced Average Variance Estimation """
        
        #mod_auto = sudire('save', center_data= True, scale_data=True,n_components=self.struct_dim)
        #mod_auto.fit(self.x_train.values, self.y_train.values)
        res_save = SAVE(self.x_train.values, self.y_train.values,6,self.struct_dim,'continuous',False,False)
        test_ans = 1.4142135623730943
        np.testing.assert_almost_equal(np.linalg.norm(res_save),test_ans,decimal=14)
        
    def test_dr(self):
        """ Tests Directional Regression """
        
        #mod_auto = sudire('dr', center_data= True, scale_data=True,n_components=self.struct_dim)
        #mod_auto.fit(self.x_train.values, self.y_train.values)
        res_dr = DR(self.x_train.values, self.y_train.values,6,self.struct_dim,'continuous',False,False)
        test_ans = 1.4142135623730934
        np.testing.assert_almost_equal(np.linalg.norm(res_dr),test_ans,decimal=14)
        
    def test_iht(self):
        """ Tests Iterative Hessian Transformations """
        
        #mod_auto = sudire('iht', center_data= True, scale_data=True,n_components=self.struct_dim)
        #mod_auto.fit(self.x_train.values, self.y_train.values)
        res_iht = IHT(self.x_train.values, self.y_train.values,self.struct_dim,False,False)
        test_ans = 1.4142135623730934
        np.testing.assert_almost_equal(np.linalg.norm(res_iht),test_ans,decimal=14)
        
    def test_phd(self):
        """ Tests Principal Hessian Directions """
        
        #mod_auto = sudire('phd', center_data= True, scale_data=True,n_components=self.struct_dim)
        #mod_auto.fit(self.x_train.values, self.y_train.values)
        res_phd = PHD(self.x_train.values, self.y_train.values,self.struct_dim,False,False)
        test_ans = 1.4142135623730894
        np.testing.assert_almost_equal(np.linalg.norm(res_phd),test_ans,decimal=14)
        
        
#    def test_dcov(self):  
#        """ Test DCOV based SDR"""
#        
#        mod_auto = sudire('dcov-sdr', center_data= True, scale_data=True,n_components=self.struct_dim)
#        mod_auto.fit(self.x_train.values, self.y_train.values)
#        test_ans = 1.1985000652583924
#        np.testing.assert_almost_equal(np.linalg.norm(mod_auto.x_loadings_),test_ans,decimal=5)
#        
#        
#    def test_mdd(self):
#        
#        """ Test MDD based SDR"""
#        mod_auto = sudire('mdd-sdr', center_data= True, scale_data=True,n_components=self.struct_dim)
#        mod_auto.fit(self.x_train.values, self.y_train.values)
#        test_ans = 0.3793912951554523
#        np.testing.assert_almost_equal(np.linalg.norm(mod_auto.x_loadings_),test_ans,decimal=5)
    
        
    
        
        
        
        
        
        
        
        
        
        
if __name__ =='__main__':
    unittest.main()
        
 
        
        



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Mar 25 09:02:05 2020

@author: Sven Serneels, Ponalytics. 

Code for radial transform functions largely adapted from 
R code by Jakob Raymaekers

"""

import numpy as np

def quad(dd, p, n): 
    """
    Computes the quadratic radial function
    args:
        dd: vector of distances
        p: number of variables in original data
        n: number of rows in original data
    returns:
        xi: radial function
    """
    d_hmed = np.sort(dd,axis=0)[int(np.floor((n + p + 1) / 2))-1][0]
    idx = np.where(dd > d_hmed)[0]
    xi = np.ones((n,1))
    xi[idx] = (1 / np.square(dd[idx])) * (d_hmed**2)
    return(xi) 

def ss(dd, p,*args,prec=1e-10):

    """
    Computes the spatial sign radial function
    args:
      dd: vector of distances
      p: dimension of original data
      *args flag to be able to pass on n - has no effect
    returns:
      xi: radial function
    """
    dd = np.maximum(dd,prec)
    xi = 1 / dd
    return(xi)

def winsor(dd, p, n) :
    """
    Computes the Winsor radial function
    args:
      dd: vector of distances
      p: number of variables in original data
      n: number of rows in original data
    returns:
      xi: radial function
    """ 
    d_hmed  = np.sort(dd,axis=0)[int(np.floor((n + p + 1) / 2))-1][0]
    idx = np.where(dd > d_hmed)[0]
    xi = np.ones((n,1))
    xi[idx] = (1 / dd[idx]) * d_hmed
    return(xi)   
    
def ball(dd, p, n): 
    
    """
    Computes the Ball radial function
    args:
      dd: vector of distances
      p: number of variables in original data
      n: number of rows in original data
    returns:
      xi: radial function
    """
    
    dWH = np.power(dd,2/3) 
    dWH_hmed = np.sort(dWH,axis=0)[int(np.floor((n + p + 1) / 2))-1][0]
    d_hmed = np.power(dWH_hmed,3/2)
    idx = np.where(dd > d_hmed)[0]
    xi = np.ones((n,1))
    xi[idx] =  0
    return(xi)


def shell(dd, p, n) :
    """
    Computes the Shell radial function
    args:
      dd: vector of distances
      p: number of variables in original data
      n: number of rows in original data
    returns:
      xi: radial function
    """
    
    dWH = np.power(dd,2/3) 
    dWH_hmed = np.sort(dWH,axis=0)[int(np.floor((n + p + 1) / 2))-1][0]
    dWH_hmad = np.sort(np.abs(dWH - dWH_hmed),axis=0)[int(np.floor((n + p + 1) / 2))-1][0]
    cutoff1 = np.power(np.maximum(0, dWH_hmed - dWH_hmad),3/2)
    cutoff2 = np.power(dWH_hmed + dWH_hmad,3/2)
    idxlow = np.where(dd < cutoff1)[0] 
    idxhigh = np.where(dd > cutoff2)[0] 
    xi = np.ones((n,1))
    xi[idxlow] = 0
    xi[idxhigh] = 0
    return(xi)


def linear_redescending(dd, p,n): 
    """
    # Computes the Linear redescending radial function
    args:
      dd: vector of distances
      p: number of variables in original data
      n: number of rows in original data
    returns:
      xi: radial function
    """
    
    dWH = np.power(dd,2/3) 
    dWH_hmed = np.sort(dWH,axis=0)[int(np.floor((n + p + 1) / 2))-1][0]
    dWH_hmad = np.sort(np.abs(dWH - dWH_hmed),axis=0)[int(np.floor((n + p + 1) / 2))-1][0]
    d_hmed = dWH_hmed**(3/2)
    cutoff = (dWH_hmed + 1.4826 * dWH_hmad)**(3/2)
    idxmid = np.where(np.logical_and(dd > d_hmed,dd <= cutoff))[0]
    idxhigh = np.where(dd > cutoff)[0]
    xi = np.ones((n,1))
    xi[idxmid] = 1 - (dd[idxmid,:] - d_hmed) / (cutoff - d_hmed)
    xi[idxhigh] = 0
    return(xi)
    

def _norms(X,**kwargs):
    """
    Casewise norms of a matrix
    """
    return(np.linalg.norm(X,axis=1,keepdims=True,**kwargs))
    
    
def _gsspp(X,p,n,fun=ss):
    """
    Generalized Spatial Sign Pre-Processing for Centred Data
    """
    return(np.multiply(X,fun(_norms(X),p,n)))
        
def _spatial_sign(X,**kwargs):
    """
    Spatial Sign Pre-Processing for Centred Data
    """
    return(X/_norms(X))
    

    
    

    


# -*- coding: utf-8 -*-
"""
Created on Sat Apr 11 17:22:09 2020

@author: Emmanuel Jordy Menvouta
"""



__name__ = "utils"
__author__ = "Emmanuel Jordy Menvouta and Sven Serneels"
__license__ = "MIT"
__version__ = "0.0.2"
__date__ = "2020-12-24"
